'''
>>> from cryptonita.conv import B           # byexample: +timeout=10
>>> from cryptonita.stats.kasiski import *
'''

from cryptonita.conv import B
from bisect import bisect_left

from operator import itemgetter
from collections import defaultdict
import itertools

from collections import Counter

# References:
# Automating the Cracking of Simple Ciphers, Matthew C. Berntsen


def as_ngram_repeated_positions(s, n, allow_overlapping=True):
    ''' Given a string of bytes returns a sorted list of (position, id) tuples.

        >>> s = B(b'ABCDBCDABCDBC')
        >>> pos_sorted = as_ngram_repeated_positions(s, n=3)

        The sorted list is a list of tuples which map the position
        of each non-unique ngram and the ngram's id, all sorted by
        position.

        In this case, the ngram of length 3:

        >>> pos_sorted
        [(0, 1), (1, 2), (2, 3), (3, 4), (4, 2), (7, 1), (8, 2), (9, 3), (10, 4)]

        In other words at position 0 there is a non-unique ngram 1, at position
        1 there is another non-unique ngram 2, .... at position 4 there is
        the same non-unique ngram 2 found at the position 1, ...

        Storing the ngrams by value (substrings) is expensive. Instead we
        use an unique identifier for each ngram (and integer).

        If two positions have the same id, it means that both point to the
        same ngram:

        >>> [p for p, id in pos_sorted if id == 4]
        [3, 10]

        In this case if we access to those two positions we will see the same
        substring:

        >>> s[3:3+3], s[10:10+3]
        ('DBC', 'DBC')

        By definition of a ngram, we will return overlapping positions:

        >>> s = B(b'ABABABAB')
        >>> as_ngram_repeated_positions(s, n=4)
        [(0, 1), (1, 2), (2, 1), (3, 2), (4, 1)]

        Note how the ngram number 1 (ABAB) that was found in the position 0
        (pair (0,1)) was also found in the position 2 (pair (2,1)).

        Currently NOT supported but it should be possible to filter the ngrams
        that overlaps with themselves.
        This may also remove the whole ngrams if the resulting becomes unique.

        In the following example we removed the overlapping ngrams of ABAB
        and BABA. In the latter the ngram BABA disappeared:

        >>> as_ngram_repeated_positions(s, n=4, allow_overlapping=False)    # byexample: +skip
        [(0, 1), (4, 1)]

        Again, NOT supported yet.
        '''

    # Map each ngram to its own identifier and add to the pos_sorted list
    # the positions and ids in order.
    # During the scanning, count how many ngrams we see of each ngram type.
    #
    # Assuming a O(1) hash implementation, this is Time/Space O(n)
    id_of_ngram = {0: 0}
    pos_sorted = []
    ngram_cnt_by_id = defaultdict(
        int, [(0, 0)]
    )  # id==0 is special with a count of 0 always
    for pos, ngram in enumerate(s.ngrams(n)):
        id = id_of_ngram.setdefault(ngram, len(id_of_ngram))

        pos_sorted.append((pos, id))
        ngram_cnt_by_id[
            id] += 1  # because the ids goes from 1 to N we could use an array

    # Filter any unique ngram (count of 0)
    # This is Time/Space O(n)
    pos_sorted = [(p, id) for p, id in pos_sorted if ngram_cnt_by_id[id] > 1]

    return pos_sorted


def merge_overlaping(pos_sorted):
    '''
        Given a list of positions and ngram identities sorted
        by the position (see as_ngram_repeated_positions)
        where the ngrams are of length N, create another
        pos_sorted list for ngrams of length N+1
        discarding any unique ngram (we are interested in the repeated ones)

        The idea is that two ngrams G1 and G2 of N bytes at positions
        P1 and P2 can be merged if they satisfy P1 + 1 == P2 (that does
        not imply that the ngram built is repeated, but it is a precondition)

        >>> #           10--\ 11  /--12
        >>> # positions      \ | /
        >>> #       0123456789|||
        >>> s = B(b'ABCDBCDABCDBC')
        >>> l3_positions = as_ngram_repeated_positions(s, n=3)
        >>> [p for p, _ in l3_positions]
        [0, 1, 2, 3, 4, 7, 8, 9, 10]

        The result can be interpreted as at the positions 0 to 10 except 5 and 6
        there are a 3-ngram that are repeated.

        >>> l4_positions = merge_overlaping(l3_positions)
        >>> [p for p, _ in l4_positions]
        [0, 1, 2, 7, 8, 9]

        Now, the positions were are 4-ngrams repeated is smaller: 0 to 2
        and 7 to 9.

        >>> l5_positions = merge_overlaping(l4_positions)
        >>> [p for p, _ in l5_positions]
        [0, 1, 7, 8]

        If you want to recover the ngrams you just need
        to use the positions to get the substrings

        >>> [s[p:p+5] for p, _ in l5_positions]
        ['ABCDB', 'BCDBC', 'ABCDB', 'BCDBC']

        If we print the ngrams' identities we will see two
        numbers: one for ABCDB and the other for BCDBC:

        >>> [id for _, id in l5_positions]
        [1, 2, 1, 2]

        Continuing with the merging:

        >>> l6_positions = merge_overlaping(l5_positions)
        >>> [p for p, _ in l6_positions]
        [0, 7]

        >>> l7_positions = merge_overlaping(l6_positions)
        >>> l7_positions
        []

        >>> s = B(b'ABCDxABCExBCDyBCE')
        >>> l3_positions = as_ngram_repeated_positions(s, n=3)
        >>> l4_positions = merge_overlaping(l3_positions)
        >>> [p for p, _ in l4_positions]
        []
        '''

    # For each position P1 see if there is a P2 such as P1 + 1 == P2.
    # That would mean that in a P1 we have a ngram G1 of length N and
    # in P2 we have another ngram G2 of length N and both share N-1 bytes.
    # In particular, G1[1:] == G2[:-1].
    # This means that in P1 we have a ngram of length N+1 which value
    # is G1 + G2[-1]  (or G1[0] + G2)
    # Because G1 and G2 are ngrams that repeat, the new ngram G1 + G2[-1]
    # *may* be repeated: the algorithm produces false positives
    #
    # If for a position P1 we don't find such P2, we mark it to be deleted
    # because we didn't find any ngram of N+1 in P1. We can be *sure*
    # that a N+1 ngram will not exist: the algorithm does not produce
    # false negatives.
    #
    # This is a Time/Space O(n)
    id_of_ngram = {0: 0}
    ngram_cnt_by_id = defaultdict(int, [(0, 0)])
    for ix, (cur, nex) in enumerate(zip(pos_sorted[:-1], pos_sorted[1:])):
        pcur, id = cur
        pnex, id2 = nex

        if pcur + 1 != pnex:
            pos_sorted[ix] = (0, 0)  # delete later (index 0 is special)
        else:
            # instead of building the ngram from G1 and G2 we use
            # G1's and G2's identifiers as a temporally ngram representation
            # to map the new larger ngram to an new identifier
            # Note how (id, id2) means that the ngram was built from G1 and G2
            # *in that order* (G2 and G1 gives another ngram of course so
            # the (id, id2) tuple order is important)
            id = id_of_ngram.setdefault((id, id2), len(id_of_ngram))

            pos_sorted[ix] = (pcur, id)  # new ngram
            ngram_cnt_by_id[id] += 1

    # the last position P1 always is deleted because there is
    # not P2 such P1 + 1 == P2 *and* P1 < P2 (basically because there
    # are no more positions after P1)
    pos_sorted[-1] = (0, 0)

    # filter any position marked to be deleted (their index is 0)
    # or if it is the position to a ngram that appears once.
    # Time/Space O(n)
    pos_sorted = [(p, id) for p, id in pos_sorted if ngram_cnt_by_id[id] > 1]

    return pos_sorted


def deltas_from_positions(positions):
    '''
        Return the difference between consecutive positions.
        If the input is sorted, the deltas returned will be always
        positive.

        This computes the difference between 0 and 7:

        >>> list(deltas_from_positions([0, 7]))
        [7]

        This computes the difference between 1 and 4,
        and 4 and 8:

        >>> list(deltas_from_positions([1, 4, 8]))
        [3, 4]

        This is another example:

        >>> list(deltas_from_positions([1, 4, 7, 11]))
        [3, 3, 4]

    '''
    # Time O(n)
    return (y - x for x, y in zip(positions[:-1], positions[1:]))


def frequency_of_deltas(s, start=3, end=None):
    '''
        Return a list of frequencies of gaps between repeated n-grams.

        >>> s = B(b'ABCDBCDABCDBC')
        >>> frequency_of_deltas(s)
        [Counter({7: 3, 3: 1, 4: 1}), Counter({7: 3}), Counter({7: 2}), Counter({7: 1})]

        The first frequencies corresponds to the 3-grams: a gap between
        two repeated 3-grams of length 7 was found 4 times; of length 3
        was found once and of length of 4 was found once.

        The next element in the returned list are the frequencies
        of the gaps between 4-grams repeated. In this case we have a single
        gap of length 7 repeated 3 times.

        The next element is for gaps between 5-grams, and so on.

        A subrange of frequencies can be obtained. For example we could be
        interested in only the frequencies of the 4-ngrams and 5-ngrams only.

        It is faster to do then the following (<end> is not inclusive):

        >>> frequency_of_deltas(s, start=4, end=6)
        [Counter({7: 3}), Counter({7: 2})]
    '''
    res = []
    n = start
    pos_sorted = as_ngram_repeated_positions(s, n=n)
    while pos_sorted:
        # we group the positions by id: each group of positions
        # will have the same identifier and therefore will belong
        # to the same ngram.
        #
        # The grouping preserves the order of the positions: positions
        # of the same ngram will stay sorted.
        #
        # This makes the compute of deltas easier (see
        # deltas_from_positions)
        #
        # O(n)
        pos_grouped = defaultdict(list)
        for pos, id in pos_sorted:
            pos_grouped[id].append(pos)

        # for each group (ngram) compute the differences between
        # its positions or "gaps"
        # This is O(n)
        delta_stats = Counter()
        for positions in pos_grouped.values():
            d = deltas_from_positions(positions)
            delta_stats.update(Counter(d))

        res.append(delta_stats)

        n += 1
        if end is not None and n >= end:
            break
        pos_sorted = merge_overlaping(pos_sorted)

    return res


def sort_deltas_by_probability(deltas_freqs):
    ''' Given a deltas frequencies of deltas of ngrams of different lengths,
        sort them by probability.

        The input for the function is a list of frequencies where the first
        element are the frequencies of the ngrams of length <n>, the second
        element are of ngrams of length <n+1>, then <n+2> and so on.

        >>> deltas_freqs = [Counter({7: 3, 9: 1}), Counter({9: 1})]

        A delta repeated several times produced by larger ngrams are much less
        likely to be product of random or luck. Instead, they should be
        the product of a coincidence between the plaintext and the key leaked
        in the ciphertext of a repeated key (Vigenere cihper).

        The function returns a list of tuples with the deltas and
        their "probability" sorted them from the less likely to the
        more likely to be random.

        For performance reasons an exponent is used instead of probability:
        larger exponent, lower probability.

        >>> list(sort_deltas_by_probability(deltas_freqs))
        [(7, 3), (9, 2), (9, 1)]

        Note: in the above example Counter({9: 1}) means that a ngram of
        length n+1 is repeated once with a distance of 9. Intuitively it should
        be less likely to be random however it is *much less* likely the
        distance of 7 that it was found 3 times even if the length of
        the ngrams is n and not n+1.

        That's why 7 is the correct answer and not 9.
        '''

    res = (
        (delta, freq * n) for n, freqs in enumerate(deltas_freqs, 1)
        for delta, freq in freqs.items()
    )
    return sorted(res, key=itemgetter(1), reverse=True)



'''
>>> # Convenient definitions
>>> from cryptonita import B        # byexample: +timeout=10
'''


def inv_right_shift(v, b, m):
    '''
        >>> from cryptonita.attacks.prng import inv_right_shift

        >>> y, b, m = 524889969, 11, 0x010101
        >>> v = y ^ ((y >> b) & m)

        >>> inv_right_shift(v, b, m)
        524889969

        >>> y, b, m = 0xffffffff, 4, 0xffffffff
        >>> v = y ^ ((y >> b) & m)

        >>> inv_right_shift(v, b, m)
        4294967295
    '''
    assert 0 < b < 32

    g = 0
    i = 0
    while i < 32:
        g = v ^ ((g >> b) & m)
        i += b

    return g


def inv_left_shift(v, b, m):
    '''
        >>> from cryptonita.attacks.prng import inv_left_shift

        >>> y, b, m = 524889969, 3, 0x010101
        >>> v = y ^ ((y << b) & m)

        >>> inv_left_shift(v, b, m)
        524889969

        >>> y, b, m = 0xffffffff, 4, 0xffffffff
        >>> v = y ^ ((y << b) & m)

        >>> inv_left_shift(v, b, m)
        4294967295
    '''
    assert 0 < b < 32

    g = 0
    i = 0
    while i < 32:
        g = v ^ ((g << b) & m)
        i += b

    return g


def clone_mt19937(out):
    ''' Clone the internal state of a Mersenne Twister 19937 (MT19937)
        from its output <out>.

        For MT19937 we need 624 sequential numbers at minimum to clone
        the state.

            >>> from cryptonita.attacks.prng import clone_mt19937
            >>> clone_mt19937([1, 2, 3])           # byexample: +norm-ws
            Traceback <...>
            ValueError: You need at least 624 numbers to clone the MT19937 PRNG
                        but you have only 3.
        '''

    n = 624
    if len(out) < n:
        raise ValueError(("You need at least %i numbers to clone the MT19937 PRNG" +\
                          " but you have only %i.") % (n, len(out)))

    u, d = 11, 0xffffffff
    s, b = 7, 0x9d2c5680
    t, c = 15, 0xefc60000
    l = 18

    state = []
    for y in out:
        assert isinstance(y, int)
        y = inv_right_shift(y, l, 0xffffffff)  # inv of y ^ ((y >> l) & 0)
        y = inv_left_shift(y, t, c)  # inv of y ^ ((y << t) & c)
        y = inv_left_shift(y, s, b)  # inv of y ^ ((y << s) & b)
        y = inv_right_shift(y, u, d)  # inv of y ^ ((y >> u) & d)

        state.append(y)

    g = MT19937(0)
    g.reset_state(state[:n], index=n)

    return g


# https://en.wikipedia.org/wiki/Mersenne_Twister
class MT19937:
    def __init__(self, seed):
        w, n, m, r = 32, 624, 397, 31
        a, f = 0x9908b0df, 1812433253
        W = 0xffffffff
        u, d = 11, 0xffffffff
        s, b = 7, 0x9d2c5680
        t, c = 15, 0xefc60000
        l = 18

        # Create a length n array to store the state of the generator
        self.MT = MT = []  # n size
        self.index = n + 1
        lower_mask = (1 << r) - 1
        upper_mask = (~lower_mask) & W

        # Initialize the generator from a seed
        index = n
        MT.append(seed)
        for i in range(1, n):
            MT.append((f * (MT[i - 1] ^ (MT[i - 1] >> (w - 2))) + i) & W)

        # Generate the next n values from the series x_i
        def twist():
            for i in range(n):
                x = (MT[i] & upper_mask) \
                          + (MT[(i+1) % n] & lower_mask)

                xA = x >> 1
                if (x % 2) != 0:  # lowest bit of x is 1
                    xA = xA ^ a

                MT[i] = MT[(i + m) % n] ^ xA

            self.index = 0

        # Extract a tempered value based on MT[index]
        # calling twist() every n numbers
        def extract_number():
            while 1:
                if self.index >= n:
                    twist()

                y = MT[self.index]
                y = y ^ ((y >> u) & d)
                y = y ^ ((y << s) & b)
                y = y ^ ((y << t) & c)
                y = y ^ (y >> l)

                self.index += 1
                yield y & W

        self.extract_number = extract_number

    def reset_state(self, MT, index=0):
        assert len(MT) == len(self.MT)

        self.index = index
        if not (0 <= self.index <= len(MT)):
            raise IndexError("Setting index=%i is out of range" \
                    % (self.index))

        self.MT[:] = MT

    def __iter__(self):
        return self.extract_number()



from cryptonita.helpers import are_same_length_or_fail, are_bytes_or_fail
import base64, base58
'''
>>> # Convenient definitions
>>> from cryptonita import B           # byexample: +timeout=10
>>> from cryptonita.bytestrings import MutableByteString, ImmutableByteString
'''


class SequenceMixin:
    __slots__ = ()

    def __getitem__(self, idx):
        ''' Get a byte or a slice of bytes.

            Use an index (number) to get a single byte:

                >>> a = B(b'ABCD')
                >>> a[0], a[-1]
                (65, 68)

            Use a slice object to get a slice of bytes:

                >>> a[:], a[:2], a[1:3]
                ('ABCD', 'AB', 'BC')

            The returned slices are *copies* of the original and
            are instances of ImmutableByteString:

                >>> isinstance(a[:2], ImmutableByteString)
                True

            '''
        v = super().__getitem__(idx)
        if isinstance(idx, slice):
            return type(self)(v)  # TODO, double copy?

        return v

    def splice(self, pos, sz, ins=None, ret_deleted=False):
        ''' Delete <sz> elements starting from the <pos> position.

            >>> a = B(b'ABCD')
            >>> a.splice(0, 1)  # equivalent to a[1:]
            'BCD'

            Negative positions are allowed following the same
            semantics that any indexing/slicing operation in Python

            >>> a.splice(-1, 1) # equivalent to a[:-1]
            'ABC'

            >>> a.splice(-2, 2) # equivalent to a[:-2]
            'AB'

            >>> a.splice(-2, 1) # equivalent to a[:-2] + a[-1:]
            'ABD'

            Negative sizes are not allowed.

            >>> a.splice(0, -1) # bad
            Traceback<...>
            ValueError: Negative sizes are not allowed (given -1)

            If <ret_deleted> is True, return the new spliced sequence
            and the deleted sub-sequence; otherwise return the new
            spliced only.

            >>> a.splice(-2, 2, ret_deleted=True) # equivalent to a[:-2], a[-2:]
            ('AB', 'CD')

            If <ins> is not None, insert that sequence at <pos> after
            the deletion.

            >>> a.splice(-2, 2, ins=B(b'X'), ret_deleted=True) # equivalent to a[:-2] + X, a[-2:]
            ('ABX', 'CD')

            >>> a.splice(1, 0, ins=B(b'XY')) # equivalent to a[:1] + XY + a[1:]
            'AXYBCD'
            '''
        if pos < 0:
            pos = len(self) + pos

        if sz < 0:
            raise ValueError(f"Negative sizes are not allowed (given {sz})")

        if ret_deleted:
            dels = self[pos:pos + sz]

        if not ins:
            new = self[:pos] + self[pos + sz:]
        else:
            new = self[:pos] + ins + self[pos + sz:]

        if ret_deleted:
            return new, dels
        else:
            return new

    def copy(self):
        return type(self)(super().copy())  # TODO, double copy?

    def __xor__(self, other):
        r'''
            Compute the xor between two strings of bytes with the same length.

                >>> a = B('1c0111001f010100061a024b53535009181c', encoding=16)
                >>> b = B('686974207468652062756c6c277320657965', encoding=16)

                >>> c = a ^ b
                >>> c.encode(16)
                b'746865206B696420646F6E277420706C6179'

            If the two strings have different lengths, the operation is undefined.
            Should we repeat the shorter string to match the length of the other
            or should we truncate the longer instead to match the length of the
            shorter?

            Different lengths are not allowed

                >>> short = B('\x11\x10\x00\x00')
                >>> a ^ short
                Traceback (most recent call last):
                <...>
                ValueError: Mismatch lengths. Left string has 18 bytes but right string has 4.

            There is an exception to this rule: if one of the stream is an
            infinite sequence, then (obviously) the only logical action is to
            truncate it to match the length of the shorter sequence and compute
            the xor:

                >>> plaintext = B("Burning 'em, if you ain't quick and nimble\nI go crazy when I hear a cymbal")
                >>> key = B("ICE")

                >>> (plaintext ^ key.inf()).encode(16)
                (b'0B3637272A2B2E63622C2E69692A23693A2A3C6324202D623D63343C2A26226324272765272A'
                 b'282B2F20430A652E2C652A3124333A653E2B2027630C692B20283165286326302E27282F')

        '''
        if not isinstance(other, InfiniteStream):
            are_same_length_or_fail(self, other)

        return type(self)((a ^ b for a, b in zip(self, other)))

    def __rxor__(self, other):
        return self ^ other

    def __ixor__(self, other):
        raise TypeError("You cannot modify a immutable byte string.")

    def __add__(self, other):
        ''' Concatenate two byte strings.

                >>> a = B(b'ABC')
                >>> b = B(b'DE')

                >>> a + b
                'ABCDE'

                >>> isinstance(a+b, ImmutableByteString)
                True

            Concatenate them even if they are plain bytearrays:

                >>> b'123' + a
                '123ABC'

                >>> a + b'123'
                'ABC123'

            The concatenation cannot be inplace (the string cannot be
            expended):

                >>> a += b
                Traceback <...>
                TypeError: You cannot expand an byte string.
            '''

        return type(self)(super().__add__(other))  # TODO double copy?

    def __radd__(self, other):
        return type(self)(other) + self  # TODO double copy?

    def __iadd__(self, other):
        raise TypeError("You cannot expand an byte string.")

    def __mul__(self, other):
        ''' Repeat a byte string <n> times.

                >>> a = B(b'ABC')

                >>> a * 3
                'ABCABCABC'

                >>> 2 * a
                'ABCABC'

                >>> isinstance(a*2, ImmutableByteString)
                True

            A non-positive number is allowed and returns an empty string

                >>> a * 0
                ''

                >>> a * -1
                ''

            The repeat cannot be inplace (the string cannot be
            expended):

                >>> a *= 2
                Traceback <...>
                TypeError: You cannot expand an byte string.
            '''
        return type(self)(super().__mul__(other))  # TODO double copy?

    def __rmul__(self, other):
        return type(self)(super().__mul__(other))  # TODO double copy?

    def __imul__(self, other):
        raise TypeError("You cannot expand an byte string.")

    def __lshift__(self, other):
        ''' Pushes the <other> string into self shifting all the
            bytes to the left.

            This does not increase or shrink the string length
            and returns a copy always.

                >>> s = B("ABCD")

                >>> s << B('E')
                'BCDE'

            Other than ByteString can be pushed too:

                >>> s << b'EFG'
                'DEFG'

            Pushing a much larger string will basically override
            the original content:

                >>> s << B('EFGHIJK')
                'HIJK'

            *Beware:* each push or shift involves copying all the byte
            string which can be really slow.

        '''
        n = len(other)

        if n == 0:
            return type(self)(self)

        if n > len(self):
            return type(self)(other[-len(self):])
        else:
            return self[n:] + other

    def __ilshift__(self, other):
        raise TypeError("You cannot modify a immutable byte string.")

    def encode(self, base):
        r'''
            Encode the byte string using base <x> (base 16, base 64, ...)

                >>> B('\x01\x02').encode(16)
                b'0102'

            Combine this with the decoding capabilities of B (conv.as_bytes)
            to create a (little inefficient) conversor between bases:

                >>> b16 = b"49276d206b696c6c696e6720796f757220627" + \
                ...       b"261696e206c696b65206120706f69736f6e6f" + \
                ...       b"7573206d757368726f6f6d"
                >>> B(b16, encoding=16).encode(64)
                b'SSdtIGtpbGxpbmcgeW91ciBicmFpbiBsaWtlIGEgcG9pc29ub3VzIG11c2hyb29t'

            Use decode to build a str object from the bytes.

                >>> t = B(b'ABC').decode(encoding='ascii')
                >>> isinstance(t, str)
                True
            '''

        if base == 58:
            return base58.b58encode(self)

        return getattr(base64, 'b%iencode' % base)(self)

    def pad(self, n, scheme):
        r'''Pad the byte string up to <n> bytes-boundaries using
            the padding <scheme> and return a new byte string object.

                >>> B('AAAAAAAAAAAA').pad(16, 'pkcs#7')
                'AAAAAAAAAAAA\x04\x04\x04\x04'

                >>> B('AAAAAAAAAAAABBBB').pad(16, 'pkcs#7')
                'AAAAAAAAAAAABBBB\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10'

                >>> isinstance(B('A').pad(16, 'pkcs#7'), ImmutableByteString)
                True

            The example above uses a particular scheme named 'pkcs#7' but other
            schemes are possible.

            Padding with zeros for example:

                >>> B('AAAAAAAAAAAA').pad(16, 'zeros')
                'AAAAAAAAAAAA\x00\x00\x00\x00'

            Additional tests:

                >>> for i in range(10):
                ...     print(B('A' * i).pad(8, 'pkcs#7'))
                b'\x08\x08\x08\x08\x08\x08\x08\x08'
                b'A\x07\x07\x07\x07\x07\x07\x07'
                b'AA\x06\x06\x06\x06\x06\x06'
                b'AAA\x05\x05\x05\x05\x05'
                b'AAAA\x04\x04\x04\x04'
                b'AAAAA\x03\x03\x03'
                b'AAAAAA\x02\x02'
                b'AAAAAAA\x01'
                b'AAAAAAAA\x08\x08\x08\x08\x08\x08\x08\x08'
                b'AAAAAAAAA\x07\x07\x07\x07\x07\x07\x07'

                >>> for i in range(10):
                ...     print(B('A' * i).pad(8, 'pkcs#7').unpad('pkcs#7'))
                b''
                b'A'
                b'AA'
                b'AAA'
                b'AAAA'
                b'AAAAA'
                b'AAAAAA'
                b'AAAAAAA'
                b'AAAAAAAA'
                b'AAAAAAAAA'

        '''
        if scheme == 'pkcs#7':
            assert n > 0
            npad = n - (len(self) % n)
            assert 1 <= npad <= n

            padding = type(self)([npad]) * npad

        elif scheme == 'zeros':
            assert n > 0
            npad = n - (len(self) % n)
            assert 1 <= npad <= n

            padding = type(self)([0]) * npad
        else:
            raise ValueError("Unknow padding scheme '%s'" % scheme)

        return self + padding

    def unpad(self, scheme):
        r'''
                >>> padded = B('AAAAAAAAAAAA').pad(16, 'pkcs#7')
                >>> padded.unpad('pkcs#7')
                'AAAAAAAAAAAA'

        '''
        if scheme == 'pkcs#7':
            n = self[-1]

            if n > 64 or self[-n:] != type(self)([n]) * n:
                raise ValueError(
                    "Bad padding '%s' with last byte %#x" % (scheme, n)
                )

            return self[:-n]
        else:
            raise ValueError("Unknow padding scheme '%s'" % scheme)

    def inf(self):
        return InfiniteStream(self)

    def ngrams(self, n):
        return NgramsView(self, n)

    def nblocks(self, n):
        return NblocksView(self, n)

    def join(self, *others):
        return type(self)(super().join(*others))


class MutableSequenceMixin(SequenceMixin):
    __slots__ = ()

    def __setitem__(self, idx, val):
        ''' Set a byte or a slice of bytes.

            Use an index (number) to set a single byte:

                >>> a = B(b'ABCD', mutable=True)

                >>> a[0] = b'X'
                >>> a[-1] = 69
                >>> a
                'XBCE'

            Use a slice object to set a slice of bytes (of the same
            length than the destination):

                >>> a[:] = b'1234'
                >>> a
                '1234'

                >>> a[1:3] = b'XY'
                >>> a
                '1XY4'

                >>> a[1:3] = b'to large'
                Traceback <...>
                ValueError: Mismatch lengths, setting 8 bytes into a buffer of 2 bytes length

                >>> a[::2] = b'AB'
                >>> a
                'AXB4'

            If you need to insert N bytes in a slice larger or smaller than N
            you can use isplice().

            As it is more common to use __setitem__ to replace N bytes by
            another N bytes without adding/removing bytes,
            this restriction on __setitem__ is to catch silly errors when
            the size of the input mismatches the size of the destination slice.
            '''
        if isinstance(idx, slice):
            start, stop, step = idx.indices(len(self))
            dlen = (stop - start) // step
            if len(val) != dlen:
                raise ValueError(
                    "Mismatch lengths, setting %i bytes into a buffer of %i bytes length"
                    % (len(val), dlen)
                )
        elif not isinstance(val, int):
            if len(val) != 1:
                raise ValueError(
                    "Mismatch lengths, setting %i bytes into a buffer of %i bytes length"
                    % (len(val), 1)
                )

            val = val[0]

        return super().__setitem__(idx, val)

    def isplice(self, pos, sz, ins=None, ret_deleted=False):
        ''' In-place variation of splice().

            By default returns None if <ret_deleted> is False
            or the deleted elements otherwise.

            In any case the deletion+insertion happen in-place.

            >>> a = B(b'ABCD', mutable=True)

            >>> a.isplice(1, 2) # equivalent to del a[1:3]
            >>> a
            'AD'

            Note how at difference with __setitem__, isplice allows
            to replace N elems by M elems with N != M.

            For example, replacing 1 byte by 3:

            >>> a.isplice(1, 1, B(b'XYZ'))  # equivalent to a[1:2] = XY
            >>> a
            'AXYZ'
        '''

        if pos < 0:
            pos = len(self) + pos

        if sz < 0:
            raise ValueError(f"Negative sizes are not allowed (given {sz})")

        if ret_deleted:
            dels = self[pos:pos + sz]

        if not ins:
            del self[pos:pos + sz]
        else:
            # Call super's because the __setitem__ of MutableSequenceMixin
            # will fail with an exception if len(ins) != sz
            super().__setitem__(slice(pos, pos + sz), ins)

        if ret_deleted:
            return dels

    def __ixor__(self, other):
        ''' xor <other> with <self> in place.

                >>> s = B(b'\x01\x02', mutable=True)

                >>> s ^= b'\x01\x01'
                >>> s
                '\x00\x03'

                >>> s ^= B(b'A').inf()
                >>> s
                'AB'

            See SequenceMixin.__xor__ for more about this.
        '''
        if not isinstance(other, InfiniteStream):
            are_same_length_or_fail(self, other)

        for idx in range(len(self)):
            super().__setitem__(idx, super().__getitem__(idx) ^ other[idx])

        return self

    def __ilshift__(self, other):
        ''' Shift to the left <other> into <self> in place.

                >>> s = B(b'ABC', mutable=True)

                >>> s <<= b'DE'
                >>> s
                'CDE'

                >>> s <<= b'FGHI'
                >>> s
                'GHI'

            See SequenceMixin.__lshift__ for more about this.
        '''
        n = len(other)

        if n == 0:
            return self

        if n > len(self):
            self[:] = other[-len(self):]
        else:
            self[:-n] = self[n:]
            self[-n:] = other

        return self


class ByteStatsMixin:
    __slots__ = ()

    # TODO cache me
    def count_1s(self):
        return sum(_number_of_1s_in_byte[b] for b in self)

    def hamming_distance(self, m2):
        r'''
            Return the Hamming distance between self and <m2>.

            >>> m1 = B('this is a test')
            >>> m2 = B('wokka wokka!!!')

            >>> m1.hamming_distance(m2)
            37

            The Hamming (or edit) distance is the count of how many bits
            these two string differ. It is defined for strings of the same
            lengths, so the following will fail:

            >>> m1.hamming_distance(m1 + m1)
            Traceback (most recent call last):
            <...>
            ValueError: Mismatch lengths. Left string has 14 bytes but right string has 28.

        '''

        x = self ^ m2
        return x.count_1s()


_number_of_1s_in_byte = [0] * 256
for i in range(len(_number_of_1s_in_byte)):
    _number_of_1s_in_byte[i] = (i & 1) + _number_of_1s_in_byte[i >> 1]

_number_of_1s_in_byte = tuple(_number_of_1s_in_byte)

import collections
import itertools as itools
from cryptonita.stats import entropy
import cryptonita.plots


class SequenceStatsMixin(cryptonita.plots.SequencePlotMixin):
    __slots__ = ()

    def freq(self):
        return collections.Counter(self)

    def most_common(self, n):
        elems, _ = zip(*self.freq().most_common(n))
        return elems

    def entropy(self, qk=None, base=None):
        freq = list(self.freq().values())
        return entropy(freq, qk, base)

    def iduplicates(self, distance, idx_of='second'):
        r'''Return the index of each duplicates or repeated item.

            An item is considered duplicated if another item is
            the same *and* it is <distance> items of distance.

            <distance> equal to 0 means "two consecutive items"

                >>> blocks = B('AABBAACCCCDDAADDAA').nblocks(2)
                >>> list(blocks)
                ['AA', 'BB', 'AA', 'CC', 'CC', 'DD', 'AA', 'DD', 'AA']

                >>> # the CCs
                >>> list(blocks.iduplicates(distance=0))
                [4]

                >>> # the first 2 and last 2 AAs and the DDs
                >>> list(blocks.iduplicates(distance=1))
                [2, 7, 8]

                >>> # the first 2 AAs with the last 2 AAs
                >>> list(blocks.iduplicates(distance=5))
                [6, 8]

            By default returns the indexes of the "second" item, the
            duplicated.

            But it is possible to return the index of the first:

                >>> list(blocks.iduplicates(distance=5, idx_of='first'))
                [0, 2]

            Or even of both (notice how the indexes are mixed):

                >>> list(blocks.iduplicates(distance=5, idx_of='both'))
                [0, 6, 2, 8]

        '''
        assert idx_of in ('both', 'first', 'second')

        # consecutive items if distance is 0
        # 1 item between if distance is 1, ...
        left, right = itools.tee(self)
        right = itools.islice(right, 1 + distance, None)

        # ``left`` is longer than ``right`` but it is ok: we want to drop
        # the last items from ``left`` and zip() will do that for us
        pairs = zip(left, right)

        idx = 0
        for a, b in pairs:
            if a == b:
                if idx_of in ('first', 'both'):
                    yield idx
                if idx_of in ('second', 'both'):
                    yield idx + distance + 1
            idx += 1

    def has_duplicates(self, distance):
        return next(self.iduplicates(distance), None) != None


from cryptonita.views import InfiniteStream, NgramsView, NblocksView



'''
Case-sensitive letter and bigram frequency counts
from large-scale English corpora
MICHAEL N. JONES and D. J. K. MEWHORT
Queen’s University, Kingston, Ontario, Canada

https://www.researchgate.net/publication/8090755_Case-sensitive_letter_and_bigram_frequency_counts_from_large-scale_English_corpora

Raw Case-Sensitive Single-Letter Counts
from the NYT Corpus

'''

from ..conv import B
from ..fuzzy_set import FuzzySet

# Format:
# Letter, Uppercase, Lowercase
_tmp = [
'A',280937,5263779,
'B',169474,866156,
'C',229363,1960412,
'D',129632,2369820,
'E',138443,7741842,
'F',100751,1296925,
'G',93212,1206747,
'H',123632,2955858,
'I',223312,4527332,
'J',78706,65856,
'K',46580,460788,
'L',106984,2553152,
'M',259474,1467376,
'N',205409,4535545,
'O',105700,4729266,
'P',144239,1255579,
'Q',11659,54221,
'R',146448,4137949,
'S',304971,4186210,
'T',325462,5507692,
'U',57488,1613323,
'V',31053,653370,
'W',107195,1015656,
'X',7578,123577,
'Y',94297,1062040,
'Z',5610,66423,
]

_utotal = _ltotal = 0
_en_upper_letter_freq, _en_lower_letter_freq, _en_letter_freq = {}, {}, {}
for letter, ucount, lcount in zip(_tmp[::3], _tmp[1::3], _tmp[2::3]):
    _en_upper_letter_freq[B(letter.upper())] = ucount
    _en_lower_letter_freq[B(letter.lower())] = lcount

    _utotal += ucount
    _ltotal += lcount

_en_letter_freq.update(_en_upper_letter_freq)
_en_letter_freq.update(_en_lower_letter_freq)

for freq_by_letter, total in (
        (_en_upper_letter_freq, _utotal),
        (_en_lower_letter_freq, _ltotal),
        (_en_letter_freq, _utotal + _ltotal)):
    for letter in freq_by_letter:
        freq_by_letter[letter] /= total

def en_upper_letter_freq():
    return FuzzySet(_en_upper_letter_freq)

def en_lower_letter_freq():
    return FuzzySet(_en_lower_letter_freq)

def en_letter_freq():
    return FuzzySet(_en_letter_freq)

# Resource
# http://norvig.com/mayzner.html
#
# Format
# Word Length, Count (in millons)
_tmp = [
 1, 22301.22,
 2,131293.85,
 3,152568.38,
 4,109988.33,
 5, 79589.32,
 6, 62391.21,
 7, 59052.66,
 8, 44207.29,
 9, 33006.93,
10, 22883.84,
11, 13098.06,
12,  7124.15,
13,  3850.58,
14,  1653.08,
15,   565.24,
16,   151.22,
17,    72.81,
18,    28.62,
19,     8.51,
20,     6.35,
21,     0.13,
22,     0.81,
23,     0.32,
]

_space_count = sum(c for c in _tmp[1::2])   # aka word count
_letter_count = sum(l*c for l, c in zip(_tmp[::2], _tmp[1::2]))
_space_freq = _space_count / (_space_count + _letter_count)

def etaoin_shrdlu(include_space=True, n=12):
    ''' Frequency of the lowercase letters taken from a set of
        English texts that mixed lowercase and uppercase letters.

        If include_space is True, include the space character in the
        frequency.
    '''
    s = en_lower_letter_freq()
    if include_space:
        s.scale(1-_space_freq)
        s[B(' ')] = _space_freq
        n += 1

    s.cut_off(n=n)
    return s

def tsamcin_brped(n=12):
    ''' Frequency of the uppercase letters taken from a set of English
        texts that mixed lowercase and uppercase letters.

        If you are interested in the frequency of english letters in general
        but your text is uppercase, change it to lowercase and use
        etaoin_shrdlu.

        The space character is not included.
        '''
    s = en_upper_letter_freq()
    s.cut_off(n=n)

    return s

# https://norvig.com/mayzner.html
# Most common n-grams
_tmp = '''
th     the
he     and
in     ing
er     ion
an     tio
re     ent
on     ati
at     for
en     her
nd     ter
ti     hat
es     tha
or     ere
te     ate
of     his
ed     con
is     res
it     ver
al     all
ar     ons
st     nce
to     men
nt     ith
ng     ted
se     ers
ha     pro
as     thi
ou     wit
io     are
le     ess
ve     not
co     ive
me     was
de     ect
hi     rea
ri     com
ro     eve
ic     per
ne     int
ea     est
ra     sta
ce     cti
li     ica
ch     ist
ll     ear
be     ain
ma     one
si     our
om     iti
ur     rat
'''
en_bigrams, en_trigrams = zip(*[line.split() for line in _tmp.strip().split('\n')])




import collections
import collections.abc
'''
>>> from cryptonita.helpers import bisect_left_rev, bisect_right_rev
'''


def are_same_length_or_fail(a, b):
    if len(a) != len(b):
        raise ValueError("Mismatch lengths. Left string has %i bytes but right string has %i." % \
                            (len(a), len(b)))


def are_bytes_or_fail(val, name):
    if not isinstance(val, collections.abc.ByteString):
        raise TypeError("The parameter '%s' should be a bytes-like instance but it is %s." % \
                            (name, type(val)))


def indices_from_slice_or_index(idx, l, step_must_be_one):
    if isinstance(idx, slice):
        start, stop, step = idx.indices(l)
        if step != 1 and step_must_be_one:
            raise IndexError(
                "A slice with a step/stride different of 1 is not supported"
            )

    elif isinstance(idx, int):
        # if idx is -1, the naive stop (idx+1) would be 0 which
        # is interpreted as "first block" and not "the last block"
        # which it is what the user wants
        tmp = slice(idx, idx + 1 if idx != -1 else l)
        start, stop, step = tmp.indices(l)
    else:
        raise IndexError("Invalid object as index; expected a number or a slice but %s was found" \
                % type(idx))

    # it is guaranteed that the start, stop and step are positives
    return start, stop, step


def _bisect_chk(a, lo, hi):
    if lo < 0:
        raise ValueError("Low bound must be greater or equal to zero")
    if hi is None:
        hi = len(a)
    if hi > len(a):
        raise ValueError(
            "High bound must be lesser or equal to the length of the array"
        )

    return lo, hi


def bisect_right_rev(a, x, lo=0, hi=None):
    ''' Return the index where to insert item x in list a, assuming a is sorted
        in decreasing order. (for increasing order see bisect.bisect_left)

        The return value i is such that all e in a[:i] have e >= x, and all e in
        a[i:] have e < x.  So if x already appears in the list, i points just
        beyond the rightmost x already there.

        >>> a = [5, 4, 3, 3, 2]
        >>> i = bisect_right_rev(a, 4)
        >>> a[:i], a[i:]
        ([5, 4], [3, 3, 2])

        >>> i = bisect_right_rev(a, 3)
        >>> a[:i], a[i:]
        ([5, 4, 3, 3], [2])

        Optional args lo (default 0) and hi (default len(a)) bound the
        slice of a to be searched.
        '''

    lo, hi = _bisect_chk(a, lo, hi)
    while lo < hi:
        mid = (lo + hi) // 2
        if x > a[mid]: hi = mid
        else: lo = mid + 1
    return lo


def bisect_left_rev(a, x, lo=0, hi=None):
    ''' Return the index where to insert item x in list a, assuming a is sorted
        in decreasing order. (for increasing order see bisect.bisect_left)

        The return value i is such that all e in a[:i] have e > x, and all e in
        a[i:] have e <= x.  So if x already appears in the list, i points just
        before the leftmost x already there.

        >>> a = [5, 4, 3, 3, 2]
        >>> i = bisect_left_rev(a, 4)
        >>> a[:i], a[i:]
        ([5], [4, 3, 3, 2])

        >>> i = bisect_left_rev(a, 3)
        >>> a[:i], a[i:]
        ([5, 4], [3, 3, 2])

        Optional args lo (default 0) and hi (default len(a)) bound the
        slice of a to be searched.
        '''

    lo, hi = _bisect_chk(a, lo, hi)
    while lo < hi:
        mid = (lo + hi) // 2
        if x >= a[mid]: hi = mid
        else: lo = mid + 1
    return lo



import functools
from cryptonita.helpers import bisect_left_rev, bisect_right_rev


def import_plot_libs():
    import numpy as np
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import seaborn as sns

    return np, mpl, plt, sns


# See for more plot ideas:
# https://www.machinelearningplus.com/plots/top-50-matplotlib-visualizations-the-master-plots-python/


def axes_style_decorator(*style_args, **style_kargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapped(*args, **kargs):
            with sns.axes_style(*style_args, **style_kargs):
                return func(*args, **kargs)

        return wrapped

    return decorator


class SequencePlotMixin:
    __slots__ = ()

    @axes_style_decorator('darkgrid')
    def plotfreq(self, n=32, fmin=1, fmax=None, ax=None, **kw):
        ''' Plot the count of each element in <self> in a bar-plot ordering
            the elements from the most common to the less.

            <self> must support the freq() method (see SequenceMixin)

            Limit the count to <n> elements. If <n> is None, no limit
            is enforced.

            Further pruning is done with <fmax> and <fmin>. If the frequency of
            an element is greather than <fmax> or less than <fmin>,
            the element will not be displayed.
            '''
        np, mpl, plt, sns = import_plot_libs()

        # Thanks to
        # https://stackoverflow.com/questions/33179122/seaborn-countplot-with-frequencies
        # https://stackoverflow.com/questions/48389751/how-do-i-get-matplotlib-to-order-bar-chart-in-the-same-order-as-my-list
        # https://matplotlib.org/3.1.1/gallery/statistics/barchart_demo.html
        autoplot = ax is None
        s = self

        fig, ax = plt.subplots()

        elems, freqs = zip(*s.freq().most_common(n))

        # prune
        ileft = 0 if fmax is None else bisect_left_rev(freqs, fmax)
        iright = bisect_right_rev(freqs, fmin, lo=ileft)

        elems = elems[ileft:iright]
        freqs = freqs[ileft:iright]

        # format
        elems = ["%02x" % e for e in elems] if isinstance(elems[0], int) else [
            e.fhex(8) for e in elems
        ]

        # percentiles
        fsums = np.cumsum(freqs)
        ftotal = fsums[-1]
        percentiles = [p * ftotal for p in [.05, .25, .5, .75, .95]]
        yticks = []
        ylabels = []
        for ix, f in enumerate(fsums):
            if f > percentiles[0]:
                while percentiles and f > percentiles[0]:
                    p = percentiles[0]
                    del percentiles[0]
                yticks.append(ix)
                ylabels.append(round(p / ftotal, 2))

                if not percentiles:
                    break

        ax.set_yticks(yticks)
        ax.set_yticklabels(ylabels)
        ax.set_ylabel('Percentile')

        # make twin axis: count and frequency
        ax2 = ax.twiny()

        ax.set_xlabel('Count')
        ax2.set_xlabel('Frequency [%]')

        # plot
        rects = ax.barh(elems, freqs, **kw)

        # annotate each bar with the formatted element
        for rect, el in zip(rects, elems):
            width = int(rect.get_width())

            # Shift the text to the right side of the right edge
            xloc = 5
            clr = 'black'
            align = 'left'

            # Center the text vertically in the bar
            yloc = rect.get_y() + rect.get_height() / 2
            ax.annotate(
                el,
                xy=(width, yloc),
                xytext=(xloc, 0),
                textcoords="offset points",
                ha=align,
                va='center_baseline',
                fontsize='small',
                color=clr,
                fontweight='demibold'
            )

        # fix the frequency range
        _, xmax = ax.get_xlim()
        ax2.set_xlim(0, (xmax / ftotal) * 100)

        ax2.grid(None)

        if autoplot:
            plt.show()

        return ax



from cryptonita import B
import string
'''
>>> from cryptonita import B                    # byexample: +timeout=10
>>> from cryptonita.toys.transposition_ciphers import *
'''


def railfence(ptext, key, dec=False):
    ''' Distribute the text in a number of rows or rails and combine
        them in a round-robin-like fashion to generate the ciphertext.

        For example, the combination for a key=4 (four rails) is:

        plaintext = thX_rail_fence
        rail 0      t     i     c     =>  tic
        rail 1       h   a l   n e    =>  halne
        rail 2        X r   _ e       =>  Xr_e
        rail 3         _     f        =>  _f

        Notice the zigzag movement from the rail 0 to the rail 3 and
        to the rail 0 again.

        >>> text = B('the rail fence cipher is a form of transposition cipher')
        >>> railfence(text, 3)
        'tr cir oorpt hh alfnecpe safr ftasoiincpeeie hi m nsoir'

        >>> railfence(text, 4)
        'tich mrs rhalnepesar taoinceer e iri oo nptoih fc ffsip'

        >>> railfence(text, 5)
        't i op hlfcpsa fsonceie hi m nsoir anee frtaiipercrorth'

        >>> railfence(text, 6)
        'termpihfne r socpe chioons h lepsffaineri i   rtoracati'

        >>> ctext = B('termpihfne r socpe chioons h lepsffaineri i   rtoracati')
        >>> railfence(ctext, 6, dec=True) == text
        True
    '''
    if key < 2:
        raise ValueError("Invalid key")

    if dec:

        # this is a hack, we rebuild the rail distribution
        # as if we were encrypting a dummy message of the same
        # length than the original just to know how many
        # items would be in each rail
        # i'm sure that there must be a better way of doing this...
        virtual_rails = [[] for _ in range(key)]
        virtual_rails.extend(reversed(virtual_rails[1:-1]))

        vlen = key + (key - 2)
        assert len(virtual_rails) == vlen
        for i in range(len(ptext)):
            virtual_rails[i % vlen].append(i)  # dummy/placeholders

        # now we replace our placeholders by the real ciphertext
        rails = virtual_rails[:key]
        start = 0
        for i in range(key):
            l = len(rails[i])
            rails[i] = list(ptext[start:start + l])
            start += l

        # and finally we iterate the rails like if we were encrypting
        # but instead of writing (append) we read (pop)
        virtual_rails = rails
        virtual_rails.extend(reversed(virtual_rails[1:-1]))

        tmp = []
        assert len(virtual_rails) == vlen
        for i in range(len(ptext)):
            tmp.append(virtual_rails[i % vlen].pop(0))

        return B(tmp)

    else:
        # we create the n rails (key==n)
        virtual_rails = [[] for _ in range(key)]

        # but we also extend them with n-2 "virtual" rails
        # because Python uses references,
        # the virtual_rails[1] is the same list that virtual_rails[-2],
        # the virtual_rails[2] is the same list that virtual_rails[-3],
        # and so on
        # so adding a symbol to virtual_rails[-2] is like adding the
        # symbol to virtual_rails[1]
        virtual_rails.extend(reversed(virtual_rails[1:-1]))

        vlen = key + (key - 2)
        assert len(virtual_rails) == vlen
        for i, b in enumerate(ptext):
            virtual_rails[i % vlen].append(b)

        # get the first n rails which are the real ones
        # and build the ciphertext from them
        rails = virtual_rails[:key]
        rails = (B(rail) for rail in rails)

        return B.join(rails)



from cryptonita import B
import string
from cryptonita.deps import importdep

gp = importdep('gmpy2')
'''
>>> from cryptonita import B                    # byexample: +timeout=10
>>> from cryptonita.toys.substitution_ciphers import *
'''


def atbash(ptext, alphabet=B(string.ascii_lowercase), dec=False):
    ''' Atbash encrypts each letter of the alphabet by its
        mirror letter.

        Given the following alphabet

        >>> alphabet = B('abcde')

        The mirror would be

        >>> alphabet[::-1]
        'edcba'

        Then, the encryption follows replacing each letter of
        the plaintext by the mirrored alphabet

        >>> ctext = atbash(B('aabeec'), alphabet)
        >>> ctext
        'eedaac'

        Encrypting again yields the plaintext (it is a reciprocal cipher)

        >>> atbash(ctext, alphabet)
        'aabeec'

        References:
        https://es.wikipedia.org/wiki/Atbash
        '''

    tr = dict(zip(alphabet, alphabet[::-1]))
    return substitute(ptext, tr)


def rot13(ptext, alphabet=B(string.ascii_lowercase), dec=False):
    ''' Split the alphabet in halves and swap them, using this
        as the substitution table.

        This is the same than shifting the alphabet N positions
        where N is the half of the length of the alphabet.

        When the alphabet is the ascii english letters ignoring
        the case, N is 13: that's why it is called "rot 13"

        >>> rot13(B('cryptonita'))
        'pelcgbavgn'

        >>> rot13(B('aabeec'), alphabet=B('abcde'))
        'ccdbbe'

        When the alphabet is divisible by 2, encrypting again
        decryptes the ciphertext (it is a reciprocal cipher)

        >>> rot13(B('pelcgbavgn'))
        'cryptonita'

        However, if the alphabet is not divisible by 2, the second
        encryption is "shifted" by one. To decrypt a ciphertext
        in this case pass dec=True.

        >>> rot13(B('ccdbbe'), alphabet=B('abcde'), dec=True)
        'aabeec'

        References:
        https://en.wikipedia.org/wiki/ROT13
        '''

    shift = len(alphabet) // 2
    if len(alphabet) % 2 != 0 and dec:
        shift += 1

    return caesar(ptext, alphabet, key=shift)


def caesar(ptext, alphabet=B(string.ascii_lowercase), key=3, dec=False):
    ''' Shit by 3 the alphabet and use that as the substitution
        table.

        >>> caesar(B('ave caesar'))
        'dyh fdhvdu'

        The alphabet and the shitf key can be changed:

        >>> caesar(B('ave caesar'), B('avecsr'), key=5)
        'rav ervcrs'

        To decrypt a ciphertext, encrypt again with a
        key equal -k where k is the original key (3 by default)
        or just pass the 'dec' flag

        >>> caesar(B('dyh fdhvdu'), dec=True)
        'ave caesar'

        >>> caesar(B('rav ervcrs'), B('avecsr'), key=-5)
        'ave caesar'

        References:
        https://en.wikipedia.org/wiki/Caesar_cipher
        '''
    shift = key % len(alphabet)

    enc_alphabet = alphabet[shift:] + alphabet[:shift]
    tr = dict(
        zip(enc_alphabet, alphabet) if dec else zip(alphabet, enc_alphabet)
    )

    return substitute(ptext, tr)


def affine(ptext, key, alphabet=B(string.ascii_lowercase), dec=False):
    '''
        Perform an affine transformation of the plaintext to obtain
        a ciphertext given an alphabet.

        An affine transformation requires two secret values, 'a' and 'b'
        (the key) where 'a' must be coprime of the length of the
        alphabet 'm'. This means that the only common divisor between 'a'
        and 'm' is 1.

        With this, each symbol 'x' of the alphabet is transformed as:
            a*x+b (mod m)

        Encryption then follows mapping each symbol of the alphabet to
        the transformed alphabet

        >>> affine(B('cryptonita'), key=(9, 2))
        'uzkhrypwrc'

        The inverse makes the decryption:

        >>> affine(B('uzkhrypwrc'), key=(9, 2), dec=True)
        'cryptonita'

        If 'a' is not a coprime of 'm', the encryption/decryption
        may not work

        >>> key = (4, 0)  # with m=26, a=4 are not coprime (2 divides both)
        >>> ctext = affine(B('cryptonita'), key)
        >>> affine(ctext, key, dec=True)
        'pryptonvtn'

        References:
        https://en.wikipedia.org/wiki/Affine_cipher
        '''
    a, b = key
    m = len(alphabet)

    # see each symbol in the alphabet as a number starting
    # from 0 to m-1
    # then apply the affine transformation to each number
    idx = ((a * x + b) % m for x in range(m))

    # use the indexes to build a "shuffled" version of the
    # alphabet
    enc_alphabet = (alphabet[ix] for ix in idx)

    # with that, build the translation table
    tr = dict(
        zip(enc_alphabet, alphabet) if dec else zip(alphabet, enc_alphabet)
    )

    return substitute(ptext, tr)


def substitute(ptext, tr):
    return B(tr.get(p, p) for p in ptext)



import heapq
import collections
import collections.abc
import itertools
import functools
from operator import itemgetter, mul as mul_func
'''
>>> # Convenient definitions
>>> from cryptonita import B           # byexample: +timeout=10
>>> from cryptonita.fuzzy_set import *

'''


class FuzzySet(dict, collections.abc.Set):
    # TODO talk about this!
    def __init__(self, iterable=(), pr=None, min_membership=0.0):
        ''' Create a FuzzySet from an iterable.

            If you have a dictionary where the values represent the
            probability or membership of the key you can use it directly

            >>> FuzzySet({'a': 1, 'b': 0.8, 'c': 0.4})
            {'a' -> 1.0000, 'b' -> 0.8000, 'c' -> 0.4000}

            You can cut the set setting a minimum for the membership:

            >>> FuzzySet({'a': 1, 'b': 0.8, 'c': 0.4}, min_membership=0.5)
            {'a' -> 1.0000, 'b' -> 0.8000}

            Note that the min_membership effect will persist.

            If you don't have a dict, you can use two iterables

            >>> FuzzySet(['a', 'b', 'c'], [1, 0.8, 0.4], min_membership=0.5)
            {'a' -> 1.0000, 'b' -> 0.8000}

            or you can use this python trick to unpack a list of tuples and
            get the same effect:

            >>> FuzzySet(*zip(*[('a', 1), ('b', 0.8), ('c', 0.4)]), min_membership=0.5)
            {'a' -> 1.0000, 'b' -> 0.8000}

            or, as a shortcut, set pr to 'tuple'.

            >>> FuzzySet([('a', 1), ('b', 0.8), ('c', 0.4)], pr='tuple', min_membership=0.5)
            {'a' -> 1.0000, 'b' -> 0.8000}

            You can build a fuzzy set from a set of keys. In this case you
            need to set a default probability or membership for all the keys:

            >>> FuzzySet(['a', 'b', 'c'], pr=0.8)
            {'c' -> 0.8000, 'b' -> 0.8000, 'a' -> 0.8000}

            If the default is omitted, 1.0 is used:

            >>> FuzzySet(['a', 'b', 'c'])
            {'c' -> 1.0000, 'b' -> 1.0000, 'a' -> 1.0000}

        '''

        self.min_membership = min_membership
        self._check_probability(None, min_membership, True)

        if not isinstance(iterable, dict):
            if pr is None:
                iterable = ((k, 1.0) for k in iterable)

            elif isinstance(pr, (float, int)):
                pr = float(pr)
                iterable = ((k, pr) for k in iterable)

            elif pr == 'tuple':
                iterable = iterable

            else:
                iterable = zip(iterable, pr)

            iterable = ((k, pr) for (k, pr) in iterable if pr > min_membership)

        else:
            iterable = (
                (k, pr) for (k, pr) in iterable.items() if pr > min_membership
            )

        dict.__init__(self, iterable)
        [self._check_probability(k, pr) for k, pr in self.items()]

    def __repr__(self):
        items = sorted(self.items(), key=itemgetter(1, 0), reverse=True)
        return '{' + ', '.join(
            '%s -> %0.4f' % (repr(k), p) for k, p in items
        ) + '}'

    def _check_probability(self, elem, prob, about_min_membership=False):
        if not (0 <= prob <= 1):
            if about_min_membership:
                raise ValueError(
                    "The minimum membership is %0.4f but it must be a value between 0 and 1."
                    % (prob)
                )
            else:
                raise ValueError(
                    "The membership of %s is %0.4f but it must be a value between 0 and 1."
                    % (repr(elem), prob)
                )

    def copy(self):
        return FuzzySet(self, min_membership=self.min_membership)

    def most_likely(self, n=None):
        '''
            >>> g = FuzzySet(['a', 'b', 'c'], [0.9, 0.8, 0.7])

            >>> g
            {'a' -> 0.9000, 'b' -> 0.8000, 'c' -> 0.7000}

            >>> g.most_likely()
            'a'

            >>> g.most_likely(2)
            ('a', 'b')

            '''
        nlargest = heapq.nlargest(
            1 if n == None else n, self.items(), key=itemgetter(1)
        )
        if n == None:
            return nlargest[0][0]
        else:
            return tuple(zip(*nlargest))[0]

    def __setitem__(self, elem, prob):
        r'''Add an element to the set and assign it a value of <prob>.

            >>> g = FuzzySet()

            >>> g['a'] = 0.51
            >>> g['b'] = 1
            >>> g['c'] = 0.7

            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000, 'a' -> 0.5100}

            We support updates

            >>> g['a'] = 0.3
            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000, 'a' -> 0.3000}

            The value of <prob> is the probability of the element to be in the
            set. Therefore it must be a value between 0 and 1.

            Zero values are discarded:
            >>> g['d'] = 0
            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000, 'a' -> 0.3000}

            Greater than 1 (or negative) raise an exception:
            >>> g['e'] = 42
            Traceback (most recent call last):
            <...>
            ValueError: The membership of 'e' is 42.0000 but it must be a value between 0 and 1.

            The FuzzySet can be configured to drop and not save any key lower
            than a given threshold:

            >>> g.min_membership = 0.4
            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000}

            >>> g['d'] = 0.1
            >>> g
            {'b' -> 1.0000, 'c' -> 0.7000}

            >>> g['c'] = 0.1
            >>> g
            {'b' -> 1.0000}

        '''

        self._check_probability(elem, prob)
        if prob < self.min_membership or prob == 0:
            if elem in self:
                del self[elem]

            # drop and don't save
            return

        dict.__setitem__(self, elem, prob)

    @property
    def min_membership(self):
        return self._min_membership

    @min_membership.setter
    def min_membership(self, pr):
        self._check_probability(None, pr, True)
        self._min_membership = pr

        self.cut_off(self._min_membership)

    def __getitem__(self, elem):
        return dict.get(self, elem, 0)

    def cut_off(self, n):
        r'''If <n> is an integer, drop all the elements except
            the <n> most likely ones.

            >>> g = FuzzySet(['a', 'b', 'c'], [0.9, 0.8, 0.7])

            >>> g
            {'a' -> 0.9000, 'b' -> 0.8000, 'c' -> 0.7000}

            >>> g.cut_off(2)
            >>> g
            {'a' -> 0.9000, 'b' -> 0.8000}

            If <n> is between 0 and 1, drop all the elements which
            probability is lower than <n>.

            >>> g.cut_off(0.85)
            >>> g
            {'a' -> 0.9000}

        '''

        if isinstance(n, int):
            nlargest = heapq.nlargest(n, self.items(), key=itemgetter(1))
            self.clear()
            dict.update(self, nlargest)

        else:
            for k in list(self.keys()):
                if self[k] < n:
                    del self[k]

    def scale(self, n):
        r'''Readjust the probability/membership of all the elements
            multiplying them by <n>.

            >>> g = FuzzySet(['a', 'b', 'c', 'd'], [0.4, 0.4, 0.8, 0.4])

            >>> g.scale(0.5)
            >>> g
            {'c' -> 0.4000, 'd' -> 0.2000, 'b' -> 0.2000, 'a' -> 0.2000}
        '''
        for k in self:
            self[k] *= n

    def normalize(self):
        r'''Makes the sum of all the elements to be 1.

            >>> g = FuzzySet(['a', 'b', 'c', 'd'], [0.4, 0.4, 0.8, 0.4])
            >>> g.normalize()
            >>> g
            {'c' -> 0.4000, 'd' -> 0.2000, 'b' -> 0.2000, 'a' -> 0.2000}
        '''
        s = sum(pr for pr in self.values())
        if s > 0:
            self.scale(1.0 / s)

    def update(self, other):
        r'''Perform the union of this set and the <other> and update self
            with the union.

            >>> g = FuzzySet(['a', 'b'], 0.5)

            >>> g.update(FuzzySet(['a', 'c'], [1, 0.2]))
            >>> g
            {'a' -> 1.0000, 'b' -> 0.5000, 'c' -> 0.2000}

        '''

        tmp = self.union(other)
        self.clear()
        dict.update(self, tmp)

    def intersection(self, other):
        r'''Return a fuzzy sets with elements that are present in both
            sets with the minimum of both membership.

            >>> a = FuzzySet(['a', 'b'], (0.7, 0.4))
            >>> b = FuzzySet(['a', 'b', 'c'], (1.0, 0.2, 0.1))

            >>> a.intersection(b)
            {'a' -> 0.7000, 'b' -> 0.2000}

            >>> a & b
            {'a' -> 0.7000, 'b' -> 0.2000}

        '''

        tmp = FuzzySet()
        min_set = other if len(other) < len(self) else self
        max_set = other if min_set is self else self

        for k, pr in min_set.items():
            tmp[k] = min(max_set[k], pr)

        return tmp

    def union(self, other):
        r'''Return a fuzzy sets with elements of both
            sets with the maximum of both membership.

            >>> a = FuzzySet(['a', 'b'], (0.7, 0.4))
            >>> b = FuzzySet(['a', 'b', 'c'], (1.0, 0.2, 0.1))

            >>> a.union(b)
            {'a' -> 1.0000, 'b' -> 0.4000, 'c' -> 0.1000}

            >>> a | b
            {'a' -> 1.0000, 'b' -> 0.4000, 'c' -> 0.1000}

        '''
        max_set = other if len(other) > len(self) else self
        min_set = other if max_set is self else self

        tmp = FuzzySet(max_set)
        for k, pr in min_set.items():
            tmp[k] = max(tmp[k], pr)

        return tmp

    def issubset(self, other):
        r'''Return True if self is a subset of the fuzzy set <other>.
            Be a subset means that all the elements in self
            are present in <other> and with a lesser membership
            value.

            >>> a = FuzzySet(['a', 'b'], (0.7, 0.4))
            >>> b = FuzzySet(['a', 'b', 'c'], (1.0, 0.2, 0.1))

            >>> a.issubset(b)
            False

            >>> a['b'] = 0.2
            >>> a.issubset(b)
            True

            >>> b.issuperset(a)
            True

            In other words, A is a subset of B if A union B yields B

            >>> a.union(b) == b
            True

        '''
        for k, pr in self.items():
            if pr > other[k]:
                return False

        return True

    def issuperset(self, other):
        return other.issubset(self)

    def __and__(self, other):
        return self.intersection(other)

    def __or__(self, other):
        return self.union(other)

    def __iand__(self, other):
        tmp = self.intersection(other)
        self.clear()
        dict.update(self, tmp)
        return self

    def __ior__(self, other):
        self.update(other)
        return self

    def sorted_items(self, most_likely_first=True):
        reverse = most_likely_first
        return sorted(self.items(), key=itemgetter(1, 0), reverse=reverse)

    def sorted_keys(self, most_likely_first=True):
        return (k for k, _ in self.sorted_items(most_likely_first))

    def sorted_values(self, most_likely_first=True):
        return (v for _, v in self.sorted_items(most_likely_first))

    @staticmethod
    def join(iterable, cut_off, j):
        return join_fuzzy_sets(iterable, cut_off, j)


def join_fuzzy_sets(iterable, cut_off, j):
    r'''
        Join the given fuzzy sets and return a single one.

        Combine all the elements of all the sets, in order, and
        build all the possible strings from that combination,
        joining the pieces with <j>.

        Each string it has a probability equals to the multiplication
        of the individual probabilities of each piece.

        For example:

        >>> A = FuzzySet(dict(y=0.9, a=0.6))
        >>> B = FuzzySet(dict(j=0.7, e=0.6))
        >>> C = FuzzySet(dict(s=0.4))

        >>> join_fuzzy_sets([A, B, C], cut_off=0.0, j='')
        {'yjs' -> 0.2520, 'yes' -> 0.2160, 'ajs' -> 0.1680, 'aes' -> 0.1440}

        Because the join produces a string for each possible combination, the
        size of the resulting set is equal to the product of the sizes of the
        input sets.

        To avoid blowing up the memory, the <cut_off> parameter can be used
        to filter away any resulting string that has a lesser than <cut_off>
        value.

        >>> join_fuzzy_sets([A, B, C], cut_off=0.2, j='')
        {'yjs' -> 0.2520, 'yes' -> 0.2160}

        If <cut_off> is greater or equal than 1, it will be interpreted as
        the number of the most probable strings instead of a minimum probability.

        >>> join_fuzzy_sets([A, B, C], cut_off=3, j='')
        {'yjs' -> 0.2520, 'yes' -> 0.2160, 'ajs' -> 0.1680}

        >>> join_fuzzy_sets([A, B, C], cut_off=1, j='')
        {'yjs' -> 0.2520}

    '''

    assert 0 <= cut_off

    if cut_off >= 1:
        counters = [len(fs) for fs in iterable]

        vals = [
            [(pr, idx) for pr in fs.values()]
            for idx, fs in enumerate(iterable)
        ]
        vals = sorted(sum(vals, []))

        for _, idx in vals:
            if counters[idx] == 1:
                continue  # skip it, at least 1 key of each fuzzy set must survive

            counters[idx] -= 1

            if functools.reduce(mul_func, counters) <= cut_off:
                counters[idx] += 1
                break

        sets = []
        for fs, counter in zip(iterable, counters):
            fs = fs.copy()
            fs.cut_off(counter)  # TODO optimize this
            sets.append(fs)

        upper_set = join_fuzzy_sets(sets, cut_off=0.0, j=j)
        upper_set.cut_off(cut_off)
        return upper_set

    all_possibilities = itertools.product(*(fs.items() for fs in iterable))

    def xxx(elems):
        byte_seq, probs = zip(*elems)

        seq_prob = 1
        for p in probs:
            seq_prob *= p

        return byte_seq, seq_prob

    # calculate the probability for each sequence
    tmp = (xxx(elems) for elems in all_possibilities)

    # filter out the less likely and join the remain byte sequences
    tmp = ((j.join(byte_seq), pr) for byte_seq, pr in tmp if pr >= cut_off)

    # build up the fuzzy set
    return FuzzySet(dict(tmp))


def len_join_fuzzy_sets(iterable):
    x = 1
    for k in iterable:
        x *= len(k)
    return x



def prefix_key_hash(hash_fun, key, message, *hash_args):
    ''' Compute H(k || m)'''
    m = key + message
    return hash_fun(m, *hash_args)


def posfix_key_hash(hash_fun, key, message, *hash_args):
    ''' Compute H(m || k)'''
    m = message + key
    return hash_fun(m, *hash_args)



from cryptonita.fuzzy_set import FuzzySet
from cryptonita import B
from cryptonita.helpers import are_bytes_or_fail

from itertools import product, zip_longest, islice
from operator import xor
'''
>>> # Convenient definitions
>>> from cryptonita import B           # byexample: +timeout=10
>>> from cryptonita.attacks import brute_force, freq_attack, search  # byexample: +timeout=10
>>> from cryptonita.fuzzy_set import FuzzySet  # byexample: +timeout=1
'''


def brute_force(ciphertext, score_func, key_space=1, min_score=0):
    r'''Guess what key was used to xor the <ciphertext>.
        Guessing means try every single possible key so we need to score
        each try with <score_func> to see what key is really useful.

        For example, if we know that the plaintext is a Bitmap image
        we can score a decrypted message like:

            >>> def is_bmp(m):
            ...     return 1 if m[:2] == b"BM" else 0

        because all the Bitmap images starts with the magic bytes 'BM'.

        Then, we guess the key:

            >>> ciphertext = B('\x1a\x15XXXY')
            >>> guessed_keys = brute_force(ciphertext, is_bmp)

            >>> guessed_keys
            {'X' -> 1.0000}


        Not always the guess is perfect. Because of this we return a FuzzySet
        object that each guessed key with its likehood.

        If <key_space> is an integer, the key space is all possible
        keys of <key_space> bytes.
        It is advised to keep <key_space> less than 3... may be 4.
        Also, use <min_score> to drop all the keys with score less than
        <min_score> to keep the memory usage at minumum.

            >>> ciphertext = B('\x1a\x14XYXX')
            >>> brute_force(ciphertext, is_bmp, key_space=2)    # byexample: +timeout=10
            {'XY' -> 1.0000}

        Otherwise, <key_space> must be an iterable of bytes: each string of bytes
        is a key to try (we will try all of them in the given order)

            >>> ciphertext = B('\x1a\x15XXXY')
            >>> possible_keys = (B(k) for k in ('Q', 'X', 'Z'))
            >>> brute_force(ciphertext, is_bmp, key_space=possible_keys)
            {'X' -> 1.0000}

        When <key_space> is a FuzzySet, not only each member is tested
        as a possible key to decipher the given <ciphertext> but also
        the likehood of the member scales the score.

        In this example the score of 1 (from is_bmp) is multiplied by the
        correct key's likehood 0.5 (from the FuzzySet):

            >>> ciphertext = B('\x1a\x15XXXY')
            >>> possible_keys = FuzzySet({B(k) : 0.5 for k in ('Q', 'X', 'Z')})
            >>> brute_force(ciphertext, is_bmp, key_space=possible_keys)
            {'X' -> 0.5000}


    '''
    assert 0.0 <= min_score <= 1.0
    are_bytes_or_fail(ciphertext, 'ciphertext')

    prob = {}
    if isinstance(key_space, int):
        key_space = (B(k) for k in product(range(256), repeat=key_space))

    elif isinstance(key_space, FuzzySet):
        prob = key_space

    key_and_likehood = (
        (k, score_func(ciphertext ^ k.inf()) * prob.get(k, 1))
        for k in key_space
    )
    keys = FuzzySet(key_and_likehood, pr='tuple', min_membership=min_score)
    return keys


def freq_attack(
    ciphertext, most_common_plain_ngrams, cipher_ngram_top=1, op=xor
):
    r'''Try to break the ciphering doing a frequency attack.
        The idea is that the plain text has some ngrams more frequent than
        others and that is reflected in the <ciphertext>.

        Then, we can assume that at least one of the <cipher_ngram_top> most
        common cipher ngrams is one of the <most_common_plain_ngrams> encrypted
        with the secret key.

        >>> most_common_plain_ngrams = [B(b) for b in b'etaoin shrdlu']
        >>> cipher_ngram_top = 1

        So, for all the most common plain ngrams p and cipher ngrams c we can
        propose (c ^ p) = k.

        >>> ciphertext = B('1b37373331363f78151b7f2b783431333d78397828372d363c78373e783a393b3736',
        ...                     encoding=16)

        >>> proposed_keys = freq_attack(ciphertext, most_common_plain_ngrams)

        >>> proposed_keys       # byexample: +norm-ws +geometry=24x400
        {<...>'X' -> 1.000<...>}


        Then we can use them to guess the secret key

        >>> from cryptonita.scoring import all_ascii_printable
        >>> key = brute_force(ciphertext, key_space=proposed_keys.keys(),
        ...                 score_func=all_ascii_printable).most_likely()

        And decrypt the message

        >>> ciphertext ^ key.inf()
        "Cooking MC's like a pound of bacon"

        Nice reading: http://norvig.com/mayzner.html
    '''

    # aliases
    T = cipher_ngram_top

    are_bytes_or_fail(ciphertext, 'ciphertext')

    keys = FuzzySet()

    prob = {}
    if isinstance(most_common_plain_ngrams, FuzzySet):
        prob = most_common_plain_ngrams

    # XXX despite the code supports different values of N in the same
    # run, that is, we support do an freq attack using ngrams of different
    # lengths, it is not clear how we should merge the results.
    #
    # Using ngrams of length 1, we can propose keys of 1 byte. If we use
    # ngrams of length 2, we can propose keys of 2 bytes.
    #
    # Now, if we have propose keys of 1 and 2 bytes how we merge them?
    # If the real secret key is of 1 byte, we should discard all the proposed
    # keys of 2 bytes that don't have those 2 bytes equals.
    #
    # Then, we will have a proposed key X (1 byte) and a proposed key XX (2 bytes)
    # but they are the same. So should we remove them too?
    # But if we do, we removed all the proposed keys of 2 bytes, what was the
    # point? It will only be useful if we don't find X but XX or we don't find XX
    # and we do find X.
    for N in (1, ):
        # count all the possible ngrams of N bytes of length of
        # the ciphertext
        # then, pick the T most common and therefore the most likely to be
        # the encrypted version of the most common plain ngrams
        _cipher_ngrams = ciphertext.ngrams(N).most_common(T)

        # most common plain ngrams of N bytes of length
        _plain_ngrams = filter(
            lambda ngram: len(ngram) == N, most_common_plain_ngrams
        )

        # if our hypothesis is correct, at least one of the c cipher ngrams
        # will be (p ^ k) where p is one of the p plain ngrams
        # if this is true, one of the 'proposed keys' keys will be the real
        # secret key k
        tmp = FuzzySet(
            (
                (op(c, p), prob.get(p, 1))
                for c, p in product(_cipher_ngrams, _plain_ngrams)
            ),
            pr='tuple'
        )
        keys.update(tmp)

    return keys


def correct_key(key, ciphertexts, suggester):
    ''' Use <key> to decrypt each <ciphertext> and for each plaintext
        try to correct the <key> to improve the quality of the plaintexts
        using the corrections suggested by <suggester>.

        Return a list of possible bytes, one for each key.
        '''
    corrections = [FuzzySet() for _ in range(len(key))]
    for ctext in ciphertexts:
        ptext = ctext[:len(key)] ^ key
        tmp = suggester(key, ctext, ptext)
        for sym_corrections, new_sym_corrections in zip(corrections, tmp):
            new_sym_corrections.normalize()
            sym_corrections |= new_sym_corrections
            sym_corrections.normalize()

    assert all(len(sym_corrections) >= 0 for sym_corrections in corrections)
    return corrections


def search(space, oracle, cnt=1, default=None):
    ''' Search the first value in the <space>
        that statisfy the <oracle> condition.

        >>> def is_4(i):
        ...     print("%i == 4?" % i)
        ...     return i == 4

        >>> search(range(2, 5), is_4)
        2 == 4?
        3 == 4?
        4 == 4?
        4

        The <space> can be any generator / iterator / sequence
        (anything that supports __iter__).

        From cryptonita.space you can find some specialized space
        definitions.

        For example, IntSpace is simil to Python's range() but
        support arbitrary starting points so you can explore
        a range of integers from the middle to the extremes.

        >>> from cryptonita.space import IntSpace

        >>> search(IntSpace(2, 10, start='middle'), is_4)
        6 == 4?
        5 == 4?
        7 == 4?
        4 == 4?
        4

        >>> search(IntSpace(2, 10, start=8), is_4)
        8 == 4?
        7 == 4?
        9 == 4?
        6 == 4?
        10 == 4?
        5 == 4?
        4 == 4?
        4

        >>> search(IntSpace(2, 10, start='end'), is_4)
        10 == 4?
        9 == 4?
        8 == 4?
        7 == 4?
        6 == 4?
        5 == 4?
        4 == 4?
        4

        With <cnt> you can control how many elements are returned (1 by default).
        The return is an iterator

        >>> def is_even(i):
        ...     print("is %i even?" % i)
        ...     return i % 2 == 0

        >>> list(search(range(10), is_even, cnt=3))
        is 0 even?
        is 1 even?
        is 2 even?
        is 3 even?
        is 4 even?
        [0, 2, 4]

        The special value `all` can be used to do a full exploration
        of the space:

        >>> list(search(range(10), is_even, cnt='all'))
        <...>
        [0, 2, 4, 6, 8]
    '''

    # TODO: add limits to search like max count of elements to explore
    # or time to do the search.
    # Can be the search paused and restored later? pickled/serialized?
    if cnt == 1:
        return next((s for s in space if oracle(s)), default)

    elif isinstance(cnt, int):
        if cnt <= 0:
            raise ValueError(
                f"Count of elements to search must be positive but {cnt} was found."
            )

        return islice((s for s in space if oracle(s)), cnt)

    elif isinstance(cnt, str):
        if cnt != 'all':
            raise ValueError(
                f"Unexpected count '{cnt}'. Accepted values are: 'all'."
            )

        return (s for s in space if oracle(s))



import importlib


class _deps_null_module:
    def __init__(self, dependency):
        self.dependency = dependency

    def __getattr__(self, attr):
        raise ImportError(
            "'{}' not installed. You may install it running 'pip install cryptonita[full]'."
            .format(self.dependency)
        )


def importdep(name):
    try:
        return importlib.import_module(name)
    except ImportError:
        return _deps_null_module(name)



'''
>>> # Convenient definitions
>>> from cryptonita import B           # byexample: +timeout=10
'''

import math
import itertools
from cryptonita.helpers import indices_from_slice_or_index
from cryptonita.conv import as_bytes


class InfiniteStream:
    __slots__ = ('base', )

    def __init__(self, base):
        self.base = base

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            n = len(self.base)
            start, stop, step = idx.start, idx.stop, idx.step
            start = start or 0
            step = step or 1

            return as_bytes(self.base[i % n] for i in range(start, stop, step))

        return self.base[idx % len(self.base)]

    def __iter__(self):
        r'''
            Return an iterator of this infinite stream of bytes.
            Basically we take a finite sequence and we repeat it
            infinite times:

                >>> s = B(b'\x01\x02').inf()

                >>> i = iter(s)
                >>> next(i), next(i), next(i), next(i)
                (1, 2, 1, 2)

            '''
        if len(self.base) <= 2048:
            return itertools.cycle(self.base)
        else:
            return self._iter_long_seq()

    def _iter_long_seq(self):
        it = iter(self.base)
        while True:
            try:
                yield next(it)
            except StopIteration:
                it = iter(self.base)

    def __repr__(self):
        '''
                >>> B(b'ABCD').inf()
                ..'ABCD'..
        '''
        return '..' + repr(self.base) + '..'

    def __len__(self):
        raise TypeError(
            "The length of an endless stream is undefined. You may think it is infinity."
        )


from cryptonita.mixins import SequenceStatsMixin


class NgramsView(SequenceStatsMixin):
    ''' N-grams view of a byte string.

        A n-gram is a sequence of n bytes that overlaps
        each other.

            >>> s = B('ABCDE')
            >>> ngrams = s.ngrams(3)

            >>> ngrams
            ['ABC', 'BCD', 'CDE']

        Shorter and larger sequences are ok

            >>> B('ABC').ngrams(3)
            ['ABC']

            >>> B('ABCDEFGHI').ngrams(3)
            ['ABC', 'BCD', 'CDE', 'DEF', 'EFG', 'FGH', 'GHI']

        But too short aren't

            >>> B('AB').ngrams(3)
            Traceback <...>
            <...>
            ValueError: The byte string has only 2 bytes. It is not possible to create even a single ngram of length 3.

        This also support some basic statistics:

            >>> ngrams = B('ABABABC').ngrams(2)
            >>> ngrams.freq()
            Counter({'AB': 3, 'BA': 2, 'BC': 1})

            >>> ngrams.most_common(n=2)
            ('AB', 'BA')

            >>> ngrams.entropy()
            1.0114042647073<...>

    '''
    __slots__ = ('base', 'n')

    def __init__(self, base, n):
        if n <= 0:
            raise ValueError("The size of a ngram cannot be zero or negative.")

        if len(base) < n:
            raise ValueError(
                "The byte string has only %i bytes. It is not possible to create even a single ngram of length %i."
                % (len(base), n)
            )

        self.base = base
        self.n = n

    def __iter__(self):
        n = self.n
        return (self.base[i:i + n] for i in range(len(self)))

    def __getitem__(self, idx):
        ''' Get a ngram or a range of ngrams

            >>> s = B(b'ABCDEF').ngrams(3)
            >>> s
            ['ABC', 'BCD', 'CDE', 'DEF']

            >>> s[-4], s[-3], s[-2], s[-1], s[0], s[1], s[2], s[3]
            ('ABC', 'BCD', 'CDE', 'DEF', 'ABC', 'BCD', 'CDE', 'DEF')

            >>> s[:2], s[-3:-1]
            (['ABC', 'BCD'], ['BCD', 'CDE'])

            >>> s[:]
            ['ABC', 'BCD', 'CDE', 'DEF']

            '''
        base_slice = self._base_slice_from_ngram_slice(idx)
        if isinstance(idx, slice):
            return self.base[base_slice].ngrams(self.n)

        return self.base[base_slice]

    def __len__(self):
        return len(self.base) - self.n + 1

    def _base_slice_from_ngram_slice(self, idx):
        n = self.n
        start, stop, step = indices_from_slice_or_index(
            idx, len(self), step_must_be_one=True
        )
        return slice(start, stop + n - 1, 1)

    def __repr__(self):
        return repr(list(self))


class NblocksView(SequenceStatsMixin):
    ''' N-blocks view of a byte string.

        A n-block is a sequence of n bytes that do not overlap
        each other.

            >>> s = B('ABCDE')
            >>> nblocks = s.nblocks(3)

            >>> nblocks
            ['ABC', 'DE']

        This also support some basic statistics:

            >>> nblocks = B('ABABCABAABBA').nblocks(2)
            >>> nblocks.freq()
            Counter({'AB': 3, 'BA': 2, 'CA': 1})

            >>> nblocks.most_common(n=2)
            ('AB', 'BA')

            >>> nblocks.entropy()
            1.0114042647073<...>

    '''
    __slots__ = ('base', 'bz')

    def __init__(self, base, block_size):
        if block_size <= 0:
            raise ValueError("The size of a block cannot be zero or negative.")
        self.base = base
        self.bz = block_size

    def __iter__(self):
        bz = self.bz
        return (self.base[i * bz:(i + 1) * bz] for i in range(len(self)))

    def __getitem__(self, idx):
        ''' Get a particular block:

            >>> nblocks = B('ABCDEFGHIJKL').nblocks(4)
            >>> nblocks[-3], nblocks[-1], nblocks[0], nblocks[1], nblocks[2]
            ('ABCD', 'IJKL', 'ABCD', 'EFGH', 'IJKL')

            >>> nblocks
            ['ABCD', 'EFGH', 'IJKL']

            Slices are supported too:

            >>> nblocks[1:2]
            ['EFGH']

            >>> nblocks[0:-1]
            ['ABCD', 'EFGH']

            >>> nblocks[-1:0]
            []

            However, slices with a step or stride different of 1 are not
            supported:

                >>> nblocks[::2]
                Traceback<...>
                IndexError: A slice with a step/stride different of 1 is not supported

            '''
        base_slice = self._base_slice_from_block_slice(idx)
        if isinstance(idx, slice):
            return self.base[base_slice].nblocks(self.bz)

        return self.base[base_slice]

    def __setitem__(self, idx, val):
        ''' Modify a block or a range of blocks.

            >>> s = B(b'ABCDEFGHIJKL', mutable=True).nblocks(4)

            >>> s[0] = B(b'1234')
            >>> s[1:2] = [B(b'5678')]

            >>> s
            ['1234', '5678', 'IJKL']

            >>> s[-3:-1] = [[65, 66, 67, 68], B(b'QWER')]
            >>> s
            ['ABCD', 'QWER', 'IJKL']

            >>> s[:-1] = B(b'12345678', mutable=True).nblocks(4)
            >>> s
            ['1234', '5678', 'IJKL']

            Shorter or larger blocks will not work
            >>> s[0] = B(b'A')
            Traceback<...>
            ValueError: Mismatch lengths, setting 1 bytes into a buffer of 4 bytes length

            With the exception of the last block which may be shorter, but
            still it must have the correct length

            >>> q = B(b'ABCD', mutable=True).nblocks(3)
            >>> q[-1]
            'D'

            >>> q[-1] = B('X')
            >>> q
            ['ABC', 'X']

            In the case of a slice of blocks, the new slice must
            also have the correct length:

            >>> s[:-1] = [B(b'QWER')]
            Traceback<...>
            ValueError: Mismatch lengths, setting 1 blocks into a buffer of 2 blocks length
            '''

        base_slice = self._base_slice_from_block_slice(idx)
        if isinstance(idx, slice):
            nblock_span = (base_slice.stop - base_slice.start) // self.bz
            if nblock_span != len(val):
                raise ValueError("Mismatch lengths, setting %i blocks into a buffer of %i blocks length" \
                        % (len(val), nblock_span))

            for i, block in enumerate(val, base_slice.start // self.bz):
                self[i] = block

        else:
            self.base[base_slice] = val

        return self

    def _base_slice_from_block_slice(self, idx):
        bz = self.bz
        start, stop, step = indices_from_slice_or_index(
            idx, len(self), step_must_be_one=True
        )
        return slice(start * bz, stop * bz, 1)

    def __repr__(self):
        return repr(list(self))

    def __len__(self):
        ''' The length of the block sequence:
            >>> len(B('ABCDEFGHIJKL').nblocks(4))
            3

            If the sequence is not multiple of the block size, the last
            block will have less items:

            >>> b = B('ABCDEFGHIJKL').nblocks(5)
            >>> len(b)
            3

            >>> b
            ['ABCDE', 'FGHIJ', 'KL']

            '''
        return math.ceil(len(self.base) / self.bz)

    def copy(self):
        return self.base.copy().nblocks(self.bz)



import json
import math
import itertools
from collections import Counter

from cryptonita.deps import importdep

np = importdep('numpy')
stats = importdep('scipy.stats')
detect_langs = importdep('language.detect_langs')

from cryptonita import B
from cryptonita.helpers import are_bytes_or_fail, are_same_length_or_fail
from cryptonita.metrics import icoincidences
'''
>>> # Convenient definitions
>>> from cryptonita import B           # byexample: +timeout=10
>>> from cryptonita.scoring import *
>>> from cryptonita.scoring.freq import etaoin_shrdlu
'''


def all_ascii_printable(m):
    ''' Score with 1 if the message has only ASCII printable characters.

            >>> import string

            >>> message = B(string.printable, encoding='ascii')
            >>> all_ascii_printable(message)
            1

        Score with 0 otherwise

            >>> message = B(string.printable + "\x04", encoding='ascii')
            >>> all_ascii_printable(message)
            0

    '''
    are_bytes_or_fail(m, 'm')
    return 1 if all((32 <= b <= 126 or 9 <= b <= 13) for b in m) else 0


def all_in_alphabet(m, alphabet):
    ''' Score 1 if all the elements in the message <m> are in the <alphabet>.

        >>> alphabet = set(B(string.printable, encoding='ascii'))

        >>> message = B('hello world', encoding='ascii')
        >>> all_in_alphabet(message, alphabet)
        1

        Score with 0 otherwise

        >>> message = B('hello\x00world', encoding='ascii')
        >>> all_in_alphabet(message, alphabet)
        0

        If the <alphabet> is the sequence of all ASCII printable characters,
        use all_ascii_printable instead, it is faster.

        Use all_in_alphabet for 'custom' alphabets.
    '''

    are_bytes_or_fail(m, 'm')
    return 1 if len(set(m) - set(alphabet)) == 0 else 0


def fit_freq_score(m, expected_prob, return_p=False, significance=0.05):
    '''
        >>> expected_prob = etaoin_shrdlu()
        >>> fit_freq_score(B("These are not the droids you are looking for"),
        ...         expected_prob)
        0.5

        >>> fit_freq_score(B("7h353 4r3 n07 7h3 dr01d5 y0u 4r3 l00k1n6 f0r"),
        ...         expected_prob)
        0
    '''
    N = len(m)

    alphabet = [k for k in expected_prob]
    alphabet.sort()

    efreq = [expected_prob[k] * N for k in alphabet]

    ofreq = m.freq()
    ofreq = [ofreq.get(k, 0) for k in alphabet]

    # The "unexpected" events:
    # If the expected probabilities don't sum up to 1 (the expected freq
    # don't sum up to N), the missing prob/freq are for the "unexpected" events
    # Make sure it is a non-zero very low frequency
    _min = min(efreq) / 16
    efreq.append(max(_min, N - sum(efreq)))
    ofreq.append(N - sum(ofreq))  # the observed unexpected events can be zero

    x = np.arange(len(efreq))
    bins = [0]
    current_bin = 0
    for i in x:
        current_bin += efreq[i]
        if current_bin > 8:
            bins.append(i + 1)
            current_bin = 0
    if current_bin > 0:
        bins.append(x[-1] + 1)

    assert len(efreq) == len(ofreq) == len(x)

    ehist, bins = np.histogram(x, weights=efreq, bins=bins)
    ohist, _ = np.histogram(x, weights=ofreq, bins=bins)

    _, p = stats.chisquare(ohist, ehist)

    return p if return_p else (0 if p <= significance else 0.5)


def good_written_word_score(m, speller, word_weight_fun=len):
    wfun = (lambda w: 1) if word_weight_fun is None else word_weight_fun
    words = m.split(sep=None)
    return sum(speller.check(w) * wfun(w)
               for w in words) / sum(wfun(w) for w in words)


def good_written_word_bit_score(m, speller):
    words = m.split(sep=None)
    score = 0
    for word in words:
        if speller.check(word):
            score += 1
        else:
            suggestions = speller.suggest(word)
            bit_length = len(word) * 8

            min_hamming_distance = bit_length
            w = B(word)
            for s in (s for s in suggestions if len(s) == len(word)):
                hd = w.hamming_distance(B(s).inf())
                if hd < min_hamming_distance:
                    min_hamming_distance = hd

            score += (bit_length - min_hamming_distance) / bit_length

    return score / len(words)


def is_language(m, language):
    return detect_langs(m)[language]


def ngram_entropy_score(m, N=1):
    r'''
        Take the frequency of each ngram of <N> bytes of length from
        <m>, and estimate the entropy.

        >>> truly_random = B('3nQucdIjAbP1n8wZ7c+xs4KoRPo6U1GP3jJllAzU24rYiCo6' +\
        ...                  'ZOgWt/zq5nG0XyNKPO95dBytg90rXxGDJGzRCNhiFX9GLZgV' +\
        ...                  'NKxbbbvrZ3bBc1b6TcthBvMWYvGAkdmOeR/TDaJgd0/x5nt5' +\
        ...                  'SiZ7Aq1+n2Eood2LxCLYCiytrp8=', encoding=64)

        >>> ngram_entropy_score(truly_random, N=1)
        4.55<...>

        Lower values are associated with string less random.

        >>> no_random = B([0b001, 0b010, 0b110, 0b001] * 64)
        >>> ngram_entropy_score(no_random, N=1)
        1.039<...>

        It is quite useful to distinguish short random sequences from not-much
        random sequences.

        It is related with icoincidences, but in the practice, entropy
        is slower and it is less discriminant.

        Nota bene: the entropy is calculated as:

            S = -sum(pk * log(pk))

        where pk is the frequency of the kth ngram and log is the natural
        logarithm (base e).

    '''
    return m.ngrams(N).entropy()


def yes_no_score(m, yes_prob=0.5):
    r'''
        Count how many 1s (yes or successes) the <m> strings has (a byte
        string that it will be seing like a string of bits).

        Then, under the hypothesis of that the sequence <m> is a Bernulli
        process (a random uniform string), <yes_prob> is the hypothetical
        probability of yes or success.

        >>> truly_random = B('3nQucdIjAbP1n8wZ7c+xs4KoRPo6U1GP3jJllAzU24rYiCo6' +\
        ...                  'ZOgWt/zq5nG0XyNKPO95dBytg90rXxGDJGzRCNhiFX9GLZgV' +\
        ...                  'NKxbbbvrZ3bBc1b6TcthBvMWYvGAkdmOeR/TDaJgd0/x5nt5' +\
        ...                  'SiZ7Aq1+n2Eood2LxCLYCiytrp8=', encoding=64)

        Under this hypothetical (named the null hypothesis), return the
        probability of *reject* the hypothesis.

        >>> yes_no_score(truly_random, yes_prob=0.5)
        0.488<...>

        The value returned is the probability of reject the hypothesis or
        the p-value [1].

        Even if the p-value is really small so we cannot reject the hypothesis,
        we *cannot* say then that <m> *is* truly uniformely random;
        We *only* can say that we don't have evidence to *reject*
        the hypothesis.

        Only high values allow us to reject the hypothesis, nothing else.

        >>> no_random = B([0b001, 0b010, 0b110, 0b001] * 64)
        >>> yes_no_score(no_random, yes_prob=0.5)
        1.0

        [The p-value] "does not tell us what we want to know,
        and we so much want to know what we want to know that,
        out of desperation, we nevertheless believe that it does"
                                                        (Cohen, 1994).

        References:

        [1] p-value: https://en.wikipedia.org/wiki/P-value
        [2] binomial test: https://en.wikipedia.org/wiki/Binomial_test
    '''

    sucessess = m.count_1s()
    bits = len(m) * 8

    return 1 - stats.binom_test(sucessess, n=bits)


def ic_score(m, m2):
    return icoincidences(m, m2)


def key_length_by_hamming_distance(ciphertext, length):
    ''' Score how much likely is <length> to be the length of the key
        which was used to encrypt and obtain the ciphertext (for a xor cipher)

        This estimate is done using the hamming distance.

        If K is a random uniform distributed key space, k one key from that
        space and ki and kj two bytes of that key where they are l bytes
        separated between.

        In other words:

            ki = k[x]
            kj = k[x+l]

        For the ci and cj ciphertext bytes we have:

            pi = ci ^ ki   and   pj = cj ^ kj

        then, the hamming distance will be:

            h(ci, cj) = sum_1s((pi ^ ki) ^ (pj ^ kj))
                      = sum_1s((pi ^ pj) ^ (ki ^ kj))

        When ki != kj, then ki ^ kj is a random uniform distributed byte so
        (pi ^ pj) ^ (ki ^ kj) is also a random uniform distributed byte
        and the expected count of 1s is half (4 bits for a 8-bits byte)

        Consider now that we are in the lucky case of ki == kj,
        the hamming distance will be:

            h(ci, cj) = sum_1s((pi ^ pj))

        If we assume the plaintext space is *not* uniform distributed,
        so the count of 1s of pi ^ pj will be significantly smaller
        than 4 so the hamming distance will be smaller than 4.

        In other words, assuming that the plaintext is *not* uniform,
        pi and pj are similar.

        This will indicate that we chosen ki and kj such they are the same,
        so indeed we found a candidate length for the secret k.

        As stated before, this length estimation assumes two things:
         - K is a random uniform distributed key string
         - P is *not* a random uniform distributed plaintext string.

        If K is not a random uniform distributed key space, the hamming
        distance may not be the best discriminant.
        '''

    l = length
    if len(ciphertext) < l * 2:
        raise ValueError(
            "The ciphertext is too short to guess the key's length and it is impossible to see if a key of %i bytes could be possible."
            % l
        )

    # This computes how many bits differ between two consecutive
    # blocks of length l
    # Keep the maximum difference
    max_distance = 0
    cblocks = ciphertext.nblocks(l)
    for a, b in zip(cblocks[:-1], cblocks[1:]):
        if len(b) < l:
            # this may happen if the original ciphertext has a total length
            # not divisible by l, so the last block will have less bytes
            # In that case we discard it
            break

        max_distance = max(a.hamming_distance(b), max_distance)

    # Compute the score normalizing the distance
    return 1 - (max_distance / (l * 8))


def key_length_by_ic(ciphertext, length):
    ''' Score the possible <length> of the key that was used to encrypt
        and obtain the <ciphertext> using the Index of Coincidence (IC).

        Pick bytes from <ciphertext> that are <length> bytes of distance:
            1, l, 2l, ...

        and compute the IC.
        '''
    return icoincidences(B(ciphertext[::length]))


# (b - a + 1)**2 - 1 / 12  = (b==1, a==0) -> 3/12
# Chebyshev's_inequality

# P [ |X - u| >= kd ] <= 1/k**2    (u = mean, d = sqrt(variance), k >= 1, X random variable)
# P [ |X - u| >= a ]  <= v / a**2  (v = var, a >= 1)



'''
>>> # Convenient definitions
>>> from cryptonita import B        # byexample: +timeout=10
>>> from cryptonita.space import IntSpace
'''


class IntSpace:
    def __init__(self, a, b=None, step=1, start=None):
        ''' IntSpace defines a space of integers (range) similar to
            Python's range() but with both extremes inclusive

            Defining only one of the extremes sets the upper bound
            so the range is defined as [0 5] (both inclusive)

            >>> list(IntSpace(5))
            [0, 1, 2, 3, 4, 5]

            Defining two extremes you can set the lower an upper bounds

            >>> list(IntSpace(3, 5))
            [3, 4, 5]

            >>> list(IntSpace(5, 5))
            [5]

            >>> list(IntSpace(0))
            [0]

            Like Python's range(), you can define the step however negative
            steps are not allowed.

            >>> list(IntSpace(5, step=2))
            [0, 2, 4]

            >>> list(IntSpace(3, 5, step=2))
            [3, 5]

            >>> list(IntSpace(5, 5, step=2))
            [5]

            >>> list(IntSpace(0, step=2))
            [0]

            In contrast with Python's range(), IntSpace allows to define
            an arbitrary starting point for the iteration.

            For example you can start from the end
            (similar to Python's range(..,.., -1)):

            >>> list(IntSpace(5, start='end'))
            [5, 4, 3, 2, 1, 0]

            Another starting point is the 'middle' point:

            >>> list(IntSpace(5, start='middle'))
            [2, 1, 3, 0, 4, 5]

            Note how the iteration starts from the middle point and goes to each
            extreme alternating between numbers lower and higher.

            The reason behind this is the assumption that numbers near
            the starting points are more interesting than the ones far from it.

            The start is not restricted to those fixed labels: arbitrary numbers
            are allowed:

            >>> list(IntSpace(5, start=1))
            [1, 0, 2, 3, 4, 5]

            >>> list(IntSpace(5, start=3))
            [3, 2, 4, 1, 5, 0]

            >>> list(IntSpace(5, start=4))
            [4, 3, 5, 2, 1, 0]

            The start can be combined with the step

            >>> list(IntSpace(5, start='end', step=2))
            [5, 3, 1]

            >>> list(IntSpace(5, start='middle', step=2))
            [2, 0, 4]

            >>> list(IntSpace(5, start=1, step=2))
            [1, 3, 5]

            >>> list(IntSpace(5, start=3, step=2))
            [3, 1, 5]

            >>> list(IntSpace(5, start=4, step=2))
            [4, 2, 0]

            >>> list(IntSpace(5, start=4, step=3))
            [4, 1]

            Invalid settings:

             - Swapped lower/upper bounds

             >>> IntSpace(-2)
             Traceback<...>
             ValueError: Invalid range [0 a] with a=-2 ('a' must be non-negative)

             >>> IntSpace(5, 2)
             Traceback<...>
             ValueError: Invalid range [a b] with a=5 and b=2 ('a' must be less than or equal to 'b')

             - Non-positive step

             >>> IntSpace(5, step=0)
             Traceback<...>
             ValueError: Invalid step s=0 over range [0 5] ('s' must be positive)

             >>> IntSpace(5, step=-1)
             Traceback<...>
             ValueError: Invalid step s=-1 over range [0 5] ('s' must be positive)

             - Invalid start

             >>> IntSpace(5, start=tuple())
             Traceback<...>
             ValueError: Invalid start s=() over range [0 5]. Expected an integer or the keywords 'begin', 'middle' or 'end'

             >>> IntSpace(5, start='foo')
             Traceback<...>
             ValueError: Invalid start s=foo over range [0 5]. Expected an integer or the keywords 'begin', 'middle' or 'end'

             - Valid start but out of range

             >>> IntSpace(5, start=-1)
             Traceback<...>
             ValueError: The start s=-1 is out of the range [0 5]

             >>> IntSpace(5, start=6)
             Traceback<...>
             ValueError: The start s=6 is out of the range [0 5]
        '''

        if b is None:
            if a < 0:
                raise ValueError(
                    f"Invalid range [0 a] with a={a} ('a' must be non-negative)"
                )
            self.lo, self.hi = 0, a
        else:
            if b < a:
                raise ValueError(
                    f"Invalid range [a b] with a={a} and b={b} ('a' must be less than or equal to 'b')"
                )
            self.lo, self.hi = a, b

        assert self.lo <= self.hi

        if step <= 0:
            raise ValueError(
                f"Invalid step s={step} over range [{self.lo} {self.hi}] ('s' must be positive)"
            )

        self.step = step
        assert self.step >= 1

        if (start != None) and not isinstance(start, int) and start not in (
            "begin", "middle", "end"
        ):
            assert start is not None
            raise ValueError(
                f"Invalid start s={start} over range [{self.lo} {self.hi}]. Expected an integer or the keywords 'begin', 'middle' or 'end'"
            )

        if isinstance(start, int) and (start < self.lo or start > self.hi):
            raise ValueError(
                f"The start s={start} is out of the range [{self.lo} {self.hi}]"
            )

        if start is None or start == "begin":
            start = self.lo
        elif start == 'middle':
            start = ((self.hi + self.lo) // 2)
        elif start == 'end':
            start = self.hi
        else:
            assert isinstance(start, int)

        self.start = start
        assert self.lo <= self.start <= self.hi

    def __iter__(self):
        return self._space_generator()

    def _space_generator(self):
        lo, hi = self.lo, self.hi
        step, start = self.step, self.start

        lower = range(start - step, lo - 1, -step)
        higher = range(start, hi + 1, step)

        # yield alternating higher and lower numbers
        cnt = 0
        for h, l in zip(higher, lower):
            yield h
            yield l
            cnt += 1

        # recreate the ranges but take into account that we already
        # yielded some of them
        hi_start = start + cnt * step
        lo_start = start - (cnt + 1) * step

        # yield the remaining numbers (if any)
        if hi_start <= hi:
            # assert that we couldn't enter in the other if-case
            assert not lo_start >= lo

            higher = range(hi_start, hi + 1, step)
            for h in higher:
                yield h

        elif lo_start >= lo:
            # assert that we couldn't enter in the other if-case
            assert not hi_start <= hi

            lower = range(lo_start, lo - 1, -step)
            for l in lower:
                yield l



from cryptonita import B
from cryptonita.fuzzy_set import FuzzySet


def good_written_word_suggester(key, ciphertext, plaintext, speller):
    ''' See <plaintext> as a list of words and suggests for each
        misspelled one another word of the same length and compute
        the xor between the original word and the suggested.

        These xors are the patched needed to transform a word to the
        suggested one.

        Return a list of fuzzy set, one for each byte (not word)
        with the suggested patches.
        '''
    corrections = []
    prev_idx = 0
    for word in plaintext.split(sep=None):
        idx = plaintext.find(word, prev_idx)
        word = B(word)
        wcorrections = [FuzzySet() for _ in range(len(word))]

        if speller.check(word):
            wsuggestions = [word]
        else:
            wsuggestions = (
                B(w) for w in speller.suggest(word) if len(w) == len(word)
            )

        for wpatch in (word ^ s for s in wsuggestions):
            for sym_corrections, suggested_sym in zip(wcorrections, wpatch):
                sym_corrections[B(suggested_sym)] = 1

        for _ in range(idx - prev_idx):
            corrections.append(
                FuzzySet([B(0)])
            )  # add any whitespace before the word

        corrections.extend(wcorrections)
        prev_idx = idx + len(word)

    for _ in range(len(plaintext) - prev_idx):
        corrections.append(
            FuzzySet([B(0)])
        )  # add any whitespace after the last word

    assert len(corrections) == len(key)
    assert all(len(sym_corrections) >= 0 for sym_corrections in corrections)
    return corrections



'''
>>> from cryptonita import B           # byexample: +timeout=10
>>> from cryptonita.metrics import *
'''

from cryptonita.helpers import are_same_length_or_fail


def count_coincidences(seq1, seq2=None, aligned=False):
    ''' Count how many coincidences are between <seq1> and <seq2>.

        >>> seq1 = B(b'AABC')
        >>> seq2 = B(b'EACAB')

        A coincidence can be seen as having the same value or symbol
        at the same position in both strings:

            AABC
            EACAB
             ^

        >>> count_coincidences(seq1, seq2, aligned=True)
        1

        But we can also see a coincidence if we shift or rotate one of
        the strings.

        For all the possible shifting or alignments we will find more
        coincidences:

            AABC    AABC    AABC    AABC    AABC
            EACAB   ACABE   CABEA   ABEAC   BEACA
             ^      ^        ^^     ^          ^

        >>> count_coincidences(seq1, seq2, aligned=False)
        6

        Other elements can be compared not only byte strings, as long
        as they support iteration and .freq() method.

        >>> ng1 = seq1.ngrams(2)
        >>> ng2 = seq2.ngrams(2)

            AA AB BC      AA AB BC      AA AB BC      AA AB BC
            EA AC CA AB   AC CA AB EA   CA AB EA AC   AB EA AC CA
                                           ^^

        >>> count_coincidences(ng1, ng2, aligned=False)
        1

        If the second input is not given, calculate the count of
        coincidences if <seq1> with itself. In this case <shift> *must* be True.

            AABC     AABC     AABC     AABC
            AABC     ABCA     BCAA     CAAB
         (not count) ^                  ^

        >>> count_coincidences(seq1)
        2

        >>> count_coincidences(seq1, aligned=True)
        <...>
        ValueError: Counting the coincidences of a sequence with itself and keep it aligned will always generate the same count, the length of the sequence.

        References:
        [1] https://eldipa.github.io/book-of-gehn/articles/2019/10/04/Index-of-Coincidence.html
    '''

    if seq2 is None:
        if aligned:
            raise ValueError(
                "Counting the coincidences of a sequence with itself and keep it aligned will always generate the same count, the length of the sequence."
            )
        else:
            f1 = seq1.freq()
            f2 = f1
            offset = 1
    else:
        if aligned:
            return sum(a == b for a, b in zip(seq1, seq2))
        else:
            f1 = seq1.freq()
            f2 = seq2.freq()
            offset = 0

    count = 0
    smaller, other = (f1, f2) if len(f1) < len(f2) else (f2, f1)
    for symbol, cnt in smaller.items():
        count += (cnt - offset) * other.get(symbol, 0)

    return count


def icoincidences(seq1, seq2=None, expected=None):
    r'''
        Take the sequence <seq1> that can be:
         - a ByteString
         - a Nblocks view
         - a Ngrams view
         - any other object that support SequenceStatsMixin.freq method.

        For the given sequence see how many items are repeated:
        count the coincidences; higher
        values means that the <seq1> is less random. [1]

        Mathematically,
            Sum for all i { (ni * (ni-1)) / (N * (N-1)) }

        where <ni> is the count of items <i> and <N> is the count of all the
        items seen.
        The result is a number between 0 (uniform random) and 1 (totally biased
        not random at all).

        For example, in the given file are 327 random strings. One of them
        is actually a message encrypted by doing a xor with a key of just 1 byte.
        Therefore, it should have more coincidences than the rest of the strings.

        >>> ciphertexts = open('test/ds/4.txt', 'rb').read().strip().split(b'\n')
        >>> ciphertexts = [B(c, encoding=16) for c in ciphertexts]

        >>> scores_and_indexes = [(icoincidences(c), i) \
        ...                         for i, c in enumerate(ciphertexts)]

        Remember, higher values are better

        >>> max(scores_and_indexes)
        (0.045977<...>, 170)

        Let's check if the 170th string is our ciphered message

        >>> key = B('5') # i am doing a little cheat here to speed up the things

        >>> ciphertexts[170] ^ key.inf()
        'Now that the party is jumping\n'

        If a second sequence <seq2> is provided, <icoincidences> will
        return the Index of Coincidence considering coincidences pair by pair

        Mathematically,
            Sum for all i { 1 if seq1[i] == seq2[i] else 0 } / N

            where <N> is the length of the sequence

        >>> icoincidences(ciphertexts[43], ciphertexts[12])
        0.033<...>

        >>> icoincidences(ciphertexts[12], ciphertexts[170])
        0.0

        In this mode, icoincidences gives an idea of how similar are
        the sequences. We can see that the non-random string (170th)
        is very different than the rest.

        The result is a number between 0 (very different) and 1 (identical)
        and it is a symmetric operation.

        If <expected> is given, the returned IC will be the *relative*
        IC respect with the <expected> probability of a coincidence.

        Mathematically,
            icoincidences(seq1, expected=e) == icoincidences(seq1) / e

        See equation 13 of [2] (in the paper, the value c is our 1/<expected>)

        Typically it comes from a theoretical "expected" value (1/256 for
        a truly random sequence of bytes; 1/26 (0.0385) for a truly random
        sequence of letters of a 26-letters alphabet like the used in English)

        >>> 1/icoincidences(ciphertexts[12], expected=1/256)
        0.424<...>

        >>> icoincidences(ciphertexts[43], expected=1/256)
        0.588<...>

        >>> 1/icoincidences(ciphertexts[170], expected=1/256)
        0.0849<...>

        Note: because the relative IC can be any positive number, to
        compare those three I inverted the result if the number was greater
        than 1 making all of them between 0 and 1 and comparable.
        Notice how the relative IC of the 170th ciphertext is far from 1
        (random).

        We can do the same with the IC of two sequences:

        >>> icoincidences(ciphertexts[43], ciphertexts[12], expected=1/256)
        8.533<...>

        See the example 1 of [2] (in the paper, the value c is our 1/<expected>)

        The icoincidences function works over ngrams too:

        >>> icoincidences(ciphertexts[170].ngrams(2))
        0.0049<...>

        >>> icoincidences(ciphertexts[43].ngrams(2))
        0.0

        >>> icoincidences(ciphertexts[12].ngrams(2), ciphertexts[43].ngrams(2))
        0.0

        References:

        [1] Index of coincidence: https://en.wikipedia.org/wiki/Index_of_coincidence
        [2] The Index of Coincidence; Howard H Campaigne
        [3] Index of Coincidence Explained: https://eldipa.github.io/book-of-gehn/articles/2019/10/04/Index-of-Coincidence.html

    '''
    # The index of coincidence is meant to work over sequences of the same
    # length comparing them symbol by symbol (aligned)
    # If not seq2 is given, the index of coincidence computes the "auto" or "self"
    # version of the IC requiring obviously a non-alignment.
    if seq2 is not None:
        are_same_length_or_fail(seq1, seq2)
        aligned = True
    else:
        aligned = False

    # if seq2 is None, count_coincidences computes:
    #
    #     Sum for all i { (ni * (ni-1)) / (N * (N-1)) }
    #
    # otherwise count_coincidences computes:
    #
    #     Sum for all i { 1 if seq1[i] == seq2[i] else 0 }
    #
    counts = count_coincidences(seq1, seq2, aligned)

    # if seq2 is None, the counts are normalized by (N * (N-1))
    # otherwise by N
    norm = (len(seq1) * (len(seq1) - 1)) if seq2 is None else len(seq1)

    ic = counts / norm
    return ic if expected is None else ic / expected



import math


def _py_entropy(pk, qk=None, base=None):
    '''
        >>> from cryptonita.stats import _py_entropy    # byexample: +timeout=10

        Input normalized
        >>> _py_entropy([0.3, 0.5, 0.2])
        1.0296530140645<...>

        Input not normalized
        >>> _py_entropy([1.3, 2.5, 3.2])
        1.0382125826451<...>

        Different log base (e by default)
        >>> _py_entropy([1.3, 2.5, 3.2], base=7)
        0.5335357252487<...>

        Two inputs (Kullback-Leibler divergence)
        >>> _py_entropy([0.3, 0.5, 0.2], [0.1, 0.1, 0.8])
        0.8570437705935<...>

        Two inputs: not normalized - normalized
        >>> _py_entropy([1.3, 2.5, 3.2], [0.1, 0.1, 0.8])
        0.3137706627238<...>

        Two inputs: normalized - not normalized
        >>> _py_entropy([0.3, 0.5, 0.2], [1.1, 2.1, 3.8])
        0.24969519533828<...>

        Two inputs: different base.
        >>> _py_entropy([1.3, 2.5, 3.2], [0.1, 0.1, 0.8], base=7)
        0.1612462234580<...>

    '''
    log = math.log
    if base == None:
        base = math.e

    if qk == None:
        np = sum(pk)
        return log(np, base) - sum((p * log(p, base)) for p in pk) / np
    else:
        np = sum(pk)
        nq = sum(qk)
        return -log(np / nq, base) + sum(
            (p * log(p / q, base)) for p, q in zip(pk, qk)
        ) / np


try:
    from scipy.stats import entropy
except ImportError:
    entropy = _py_entropy



from cryptonita.deps import importdep

z3 = importdep('z3')

from cryptonita.helpers import are_same_length_or_fail, are_bytes_or_fail
'''
>>> from cryptonita import B           # byexample: +timeout=10
>>> from cryptonita.attacks.mapping import _ngrams_map
'''


def _single_ngram_map(mapping, _from, _to):
    ''' Asserts (in z3/smt terminology) that there is a mapping
        such maps <_from> to <_to>.

        Return this assertiong.
        '''
    are_same_length_or_fail(_from, _to)
    are_bytes_or_fail(_from, 'cipher-ngram')
    are_bytes_or_fail(_to, 'cipher-ngram')
    return z3.And([mapping[f] == t for f, t in zip(_from, _to)])


def _ngrams_map(possible_plain_ngrams_by_cipher_ngram, alphabet=None):
    ''' Let's assume that we are 100% sure that a particular ngram
        ('s') in the ciphertext is mapped to 'e' in the plaintext
        and that the 'c' is mapped to or 'a' or to 't'.

        In other words:

        >>> p1grams = {
        ...     b's': [b'e'],          # 100% sure
        ...     b'c': [b'a', b't'],    # 50%-50% sure
        ... }

        Let's assume also that we are 100% sure that the tri-gram
        'jds' in the ciphertext space maps to 'the' in the plaintext space

        >>> p3grams = {
        ...     b'jds': [b'the']
        ... }

        >>> alphabet = (0, 256)

        >>> p2grams = {}

        >>> possible_c2p_mappings = {}
        >>> possible_c2p_mappings.update(p1grams)
        >>> possible_c2p_mappings.update(p3grams)

        >>> #possible_plain_ngrams_by_cipher_ngram = UNION(p1grams, p2grams, p3grams)
        >>> #kmap = _ngrams_map(possible_plain_ngrams_by_cipher_ngram, alphabet)
        >>> kmap = _ngrams_map(possible_c2p_mappings, alphabet)
        '''

    if alphabet is None:
        alphabet = (0, 256)

    alph_low, alph_high = alphabet
    alph_sz = alph_high - alph_low

    kmap = z3.IntVector('kmap', alph_sz)
    assertions = []

    # Assert that every single item in the mapping (aka, every char)
    # is in the specified alphabet
    assertions.append(z3.And([km >= alph_low for km in kmap]))
    assertions.append(z3.And([km < alph_high for km in kmap]))

    for c, p_ngrams in possible_plain_ngrams_by_cipher_ngram.items():
        # Assert that the cipher ngram 'c' is mapped to one of its plain
        # ngram associated (p_grams).
        # In other words, we know that the decryption of 'c' is one of
        # the 'p' in 'p_grams' but we don't know which
        alternatives = z3.Or([_single_ngram_map(kmap, c, p) for p in p_ngrams])
        assertions.append(alternatives)

    return kmap



from cryptonita import B


def decrypt_ecb_tail(alignment, block_size, encryption_oracle, limit=None):
    align_test_block = B("A" * alignment)

    test_block = B("A" * (block_size - 1))
    align_target_block = B(test_block)  # copy
    distance = 0

    decrypted_bytes = []
    i = 0
    eof = False
    while not eof and (i < limit if limit else True):
        i += 1
        eof = True
        for b in range(256):
            b = B(b)

            # propose the following choosen plaintext:
            #
            #   |-------|-------|-------|------
            #    ....AAA AAAAAAb AAAAAA? .....
            #       |       |       |
            #       |       |  a block identical to the test block except
            #       |       |  the last byte that's unknow to us (to be decypted)
            #       |  a "test" block: a full block where the last byte
            #       |  is our guessed byte (if 'b' is equal to '?' our guess is correct)
            #    padding block for alignment purposes
            tmp = align_test_block + test_block + b + align_target_block
            c = encryption_oracle(tmp)

            # TODO i'm not resistent to possible false positive!
            if c.nblocks(block_size).has_duplicates(distance):
                # two block had collided to a <distance> blocks of distance
                # that means that out guess 'b' matched with the unknow
                # bytes '?' effectively decrypting it
                eof = False
                decrypted_bytes.append(b)

                # the test block shift to the left one byte, now that we
                # now that 'b' is the correct byte we are reserving one byte
                # on the right for the next guess
                #           |-------|
                #            AAAAAAb  -> b is guessed ok (G)
                #           |-------|
                #            AAAAAAG  -> shift the block to make room
                #           |-------|
                #            AAAAAG   -> the byte missing will be filled
                #                        with the next guess 'b'
                test_block = test_block << b

                if len(align_target_block) == 0:
                    align_target_block = B("A" * (block_size - 1))
                    distance += 1
                else:
                    align_target_block = align_target_block[:-1]

                break

    return B(b''.join(decrypted_bytes))


def decrypt_cbc_last_blk_padding_attack(cblocks, bsize, oracle):
    prev_cblock = cblocks[-2]

    x = B(range(bsize, 0, -1), mutable=True)
    x ^= prev_cblock
    for i in range(bsize - 1, -1, -1):
        prefix = prev_cblock[:i]
        padn = B(bsize - i)
        posfix = B(padn * (bsize - i - 1)) ^ x[i + 1:]

        # forge the penultimate ciphertext block
        cblocks[-2] = B(prefix + B(0) + posfix, mutable=True)
        for n in range(256):
            if prev_cblock[i] == n:
                continue

            # update the forged byte
            cblocks[-2][i] = n
            forged_ciphertext = B('').join(cblocks)

            good = oracle(forged_ciphertext)
            if good:
                x[i] = (padn ^ B(n))
                break

    cblocks[-2] = prev_cblock  # restore backup
    x ^= prev_cblock
    return x  # plain text block


def decrypt_cbc_padding_attack(ciphertext, bsize, oracle, iv=None):
    p = []
    if iv != None:
        ciphertext = iv + ciphertext

    cblocks = list(ciphertext.nblocks(bsize))

    while len(cblocks) > 1:
        p.append(decrypt_cbc_last_blk_padding_attack(cblocks, bsize, oracle))
        del cblocks[-1]

    p.reverse()
    return B('').join(p)



from cryptonita.mixins import (
    SequenceMixin, MutableSequenceMixin, ByteStatsMixin, SequenceStatsMixin
)

from cryptonita.deps import importdep

np = importdep('numpy')

from collections.abc import Iterable, Callable
'''
>>> from cryptonita import B           # byexample: +timeout=10
>>> from cryptonita.bytestrings import MutableByteString, ImmutableByteString
'''


class ImmutableByteString(
    SequenceMixin, ByteStatsMixin, SequenceStatsMixin, bytes
):
    ''' Enhanced version of an immutable byte string.

            >>> s = B(b'ABA')
            >>> isinstance(s, bytes) and type(s) != bytes
            True

        This also supports some basic statistics:

            >>> s.freq()
            Counter({65: 2, 66: 1})

            >>> s.most_common(n=1)
            (65,)

            >>> s.entropy()
            0.6365141682948<...>

            >>> list(s.iduplicates(distance=1, idx_of='first'))
            [0]

        For convenience, iterating an ImmutableByteString yields integers
        like 'bytes'

            >>> list(s)
            [65, 66, 65]

            >>> s[0]
            65

        For this reason a byte string of just 1 byte can be compared
        with an integer and both a byte and an integer can be used as keys:

            >>> one_byte = B(b'A')
            >>> (one_byte == 65), (hash(one_byte) == hash(65))
            (True, True)

            >>> isinstance(one_byte, int)
            False

            >>> one_byte in s.freq()
            True
    '''
    __slots__ = ()

    def __repr__(self):
        return super().__repr__()[1:]

    def tobytes(self):
        return self

    def __hash__(self):
        if len(self) == 1:
            return hash(self[0])
        return super().__hash__()

    def __eq__(self, other):
        if isinstance(other, int) and len(self) == 1:
            return other == self[0]
        return super().__eq__(other)

    def toarray(self):
        return np.array(tuple(self))

    def fhex(self, n=8):
        return super().hex()[:n]

    def map(self, table, table2=None, fill=-1, delete=set()):
        ''' Translate the byte string replacing each byte
            by the byte given in <table>, optionally deleting
            bytes from the <delete> set and filling the missing
            ones with <fill>.

            This is an enhanced version of bytes.translate()
            and str.translate().

            The <table> can be a bytes-like sequence of 256
            bytes and the mapping is based on position.

            Given the table:

            >>> t = B(b'ABC').pad(256, 'zeros')

            The bytes 0, 1 and 2 of the following byte string
            are translated to A, B and C respectively

            >>> s = B([0, 1, 2])
            >>> s.map(t)
            b'ABC'

            But other mapping are possible.

            If <table> is a dictionary, the mapping is by value:
            each byte is looked up in the dictionary and the value
            returned is its replacement.

            >>> s.map({0: b'A', 1: b'B', 2: b'C'})
            b'ABC'

            If two tables are given, they must have the same length
            and the mapping is by value: the ith value in the table
            of the left is mapped to the ith value of the right.

            >>> s.map([0, 1, 2], [b'A', b'B', b'C'])
            b'ABC'

            An iterator and a callable are supported too. The iterator
            must return tuples to map one item to another and the
            callable must accept one item and return its replacement

            >>> s.map(zip([0, 1, 2], [b'A', b'B', b'C']))
            b'ABC'

            >>> s.map(lambda n: [b'A', b'B', b'C'][n])
            b'ABC'

            The callables can raise LookupError or IndexError to signal
            that there is no a valid mapping for the given byte (see
            "default mapping" below). Returning None will mark the bytes
            for deletion (see "delete bytes" below).

            Note: the callable will be called 256 times to generate the
            mapping for all the possible bytes. It will *not* be called
            per each byte of <self>. For that use the standar Python
            function <map> or a for-loop.

            For all the possible mappings, inputs and outputs can be
            integers (0 <= i < 256), bytes, bytearrays or
            immutable or mutable byte strings (of length 1).

            With the exception of a single byte string of 256 values
            as a table, all the mappings are "incomplete": they define
            how to map a subset of the possible values that a byte string
            could have (256 to be precise).

            We can define a default mapping for them:

             - we can replace them by a default (fill = <default>)
             - we can leave them as they are (fill = -1)
             - we can delete them (fill = None)

            >>> s.map([(1, b'B')], fill=0x40)
            b'@B@'

            >>> s.map([(1, b'B')], fill=-1)
            b'\x00B\x02'

            >>> s.map([(1, b'B')], fill=None)
            b'B'

            Deleting some bytes is possible not only with fill=None but
            returning None from a callable

            >>> s.map(lambda x: b'B' if x == 1 else None)
            b'B'

            Also, we can define the set of bytes to delete from the
            <delete> parameter:

            >>> s.map([], delete={0x0, 0x2})
            b'\x01'

        '''
        if isinstance(table, dict):
            table = table.items()

        if table2 is not None:
            table = zip(table, table2)

        if isinstance(table, Callable):
            func = table
            table = []
            for b in range(256):
                try:
                    v = func(b)
                except LookupError:
                    v = b
                except IndexError:
                    v = b

                table.append((b, v))

        def as_int(obj):
            if isinstance(obj, (bytes, bytearray)):
                return obj[0]
            return int(obj)

        # by default do not delete anything. this is compatible with the
        # default of bytes.translate.
        delete_set = set(delete)
        if isinstance(table, Iterable) and not isinstance(table, bytes):
            if fill is None:
                # delete all of them by default
                delete_set = set(range(256))

            if fill == -1 or fill is None:
                tmp = list(range(256))
            else:
                tmp = [as_int(fill)] * 256

            for ix, v in table:
                # normalize the items as (int, int) values
                ix = as_int(ix)

                if v is None:
                    # None means delete it
                    delete_set.add(ix)
                    v = 0
                else:
                    # if it is not None means that we don't want to
                    # delete it. If the default is delete (fill is None),
                    # removing the item from the set makes sense
                    delete_set.discard(ix)
                    v = as_int(v)

                tmp[ix] = v

            table = bytes(tmp)

        deletechars = bytes(delete_set)
        return super().translate(table, deletechars)

    def tomutable(self):
        return MutableByteString(self)


class MutableByteString(MutableSequenceMixin, bytearray):
    ''' Enhanced version of a mutable byte string.

            >>> s = B(b'ABA', mutable=True)
            >>> isinstance(s, bytearray) and type(s) != bytearray
            True
    '''
    __slots__ = ()

    def __repr__(self):
        return repr(self.toimmutable())

    def toimmutable(self):
        return ImmutableByteString(self)

    def tobytes(self):
        return bytes(self)

    def toarray(self):
        return np.array(tuple(self))



# Pure python implementation of SHA256, modified to be useful
# for cryptonita
#
# From https://github.com/keanemind/Python-SHA-256
# License MIT
'''
>>> from cryptonita.toys.hashes.sha256 import sha256, sha224
'''

K = [
    0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1,
    0x923f82a4, 0xab1c5ed5, 0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
    0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174, 0xe49b69c1, 0xefbe4786,
    0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
    0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147,
    0x06ca6351, 0x14292967, 0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
    0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85, 0xa2bfe8a1, 0xa81a664b,
    0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
    0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a,
    0x5b9cca4f, 0x682e6ff3, 0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
    0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
]


def sha256(
    message,

    # Setting Initial Hash Value
    h0=0x6a09e667,
    h1=0xbb67ae85,
    h2=0x3c6ef372,
    h3=0xa54ff53a,
    h4=0x510e527f,
    h5=0x9b05688c,
    h6=0x1f83d9ab,
    h7=0x5be0cd19,
    forged_message_len=None
):
    """
    >>> sha256(b'')
    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'

    >>> sha256(b'The quick brown fox jumps over the lazy dog')
    'd7a8fbb307d7809469ca9abcb0082e4f8d5651e46d3cdb762d02d0bf37c9e592'

    >>> sha256(b'BEES')
    'd3c1c8a7366e3decde2ad0fe2d9a4fe7e0441f3d5653c0a717564ce9c3bb844c'
    """

    message = bytes(message)

    # Padding
    length = len(message) * 8  # len(message) is number of BYTES!!!
    message += b'\x80'
    while (len(message) * 8 + 64) % 512 != 0:
        message += b'\x00'

    if forged_message_len is None:
        message += length.to_bytes(8, 'big')  # pad to 8 bytes or 64 bits
    else:
        message += (forged_message_len *
                    8).to_bytes(8, 'big')  # pad to 8 bytes or 64 bits

    assert (len(message) * 8) % 512 == 0, "Padding did not complete properly!"

    # Parsing
    blocks = []  # contains 512-bit chunks of message
    for i in range(0, len(message), 64):  # 64 bytes is 512 bits
        blocks.append(message[i:i + 64])

    # SHA-256 Hash Computation
    for message_block in blocks:
        # Prepare message schedule
        message_schedule = []
        for t in range(0, 64):
            if t <= 15:
                # adds the t'th 32 bit word of the block,
                # starting from leftmost word
                # 4 bytes at a time
                message_schedule.append(
                    bytes(message_block[t * 4:(t * 4) + 4])
                )
            else:
                term1 = _sigma1(int.from_bytes(message_schedule[t - 2], 'big'))
                term2 = int.from_bytes(message_schedule[t - 7], 'big')
                term3 = _sigma0(
                    int.from_bytes(message_schedule[t - 15], 'big')
                )
                term4 = int.from_bytes(message_schedule[t - 16], 'big')

                # append a 4-byte byte object
                schedule = ((term1 + term2 + term3 + term4) %
                            2**32).to_bytes(4, 'big')
                message_schedule.append(schedule)

        assert len(message_schedule) == 64

        # Initialize working variables
        a = h0
        b = h1
        c = h2
        d = h3
        e = h4
        f = h5
        g = h6
        h = h7

        # Iterate for t=0 to 63
        for t in range(64):
            t1 = (
                (
                    h + _capsigma1(e) + _ch(e, f, g) + K[t] +
                    int.from_bytes(message_schedule[t], 'big')
                ) % 2**32
            )

            t2 = (_capsigma0(a) + _maj(a, b, c)) % 2**32

            h = g
            g = f
            f = e
            e = (d + t1) % 2**32
            d = c
            c = b
            b = a
            a = (t1 + t2) % 2**32

        # Compute intermediate hash value
        h0 = (h0 + a) % 2**32
        h1 = (h1 + b) % 2**32
        h2 = (h2 + c) % 2**32
        h3 = (h3 + d) % 2**32
        h4 = (h4 + e) % 2**32
        h5 = (h5 + f) % 2**32
        h6 = (h6 + g) % 2**32
        h7 = (h7 + h) % 2**32

    return '%08x%08x%08x%08x%08x%08x%08x%08x' % (
        h0, h1, h2, h3, h4, h5, h6, h7
    )


def sha224(
    message,

    # Setting Initial Hash Value (different than the used by SHA-256)
    h0=0xc1059ed8,
    h1=0x367cd507,
    h2=0x3070dd17,
    h3=0xf70e5939,
    h4=0x68581511,
    h5=0xffc00b31,
    h6=0x64f98fa7,
    h7=0xbefa4fa4,
    forged_message_len=None
):
    """
    >>> sha224(b'')
    'd14a028c2a3a2bc9476102bb288234c415a2b01f828ea62ac5b3e42f'

    >>> sha224(b'The quick brown fox jumps over the lazy dog')
    '730e109bd7a8a32b1cb9d9a09aa2325d2430587ddbc0c38bad911525'

    >>> sha224(b'BEES')
    '72d66b7a0ff879cc7f0f0e4c778cdd53e4a04d50466ef2650438fb6f'
    """

    h = sha256(message, h0, h1, h2, h3, h4, h5, h6, h7, forged_message_len)
    return h[:-8]


def _sigma0(num: int):
    """As defined in the specification."""
    num = (_rotate_right(num, 7) ^ _rotate_right(num, 18) ^ (num >> 3))
    return num


def _sigma1(num: int):
    """As defined in the specification."""
    num = (_rotate_right(num, 17) ^ _rotate_right(num, 19) ^ (num >> 10))
    return num


def _capsigma0(num: int):
    """As defined in the specification."""
    num = (
        _rotate_right(num, 2) ^ _rotate_right(num, 13)
        ^ _rotate_right(num, 22)
    )
    return num


def _capsigma1(num: int):
    """As defined in the specification."""
    num = (
        _rotate_right(num, 6) ^ _rotate_right(num, 11)
        ^ _rotate_right(num, 25)
    )
    return num


def _ch(x: int, y: int, z: int):
    """As defined in the specification."""
    return (x & y) ^ (~x & z)


def _maj(x: int, y: int, z: int):
    """As defined in the specification."""
    return (x & y) ^ (x & z) ^ (y & z)


def _rotate_right(num: int, shift: int, size: int = 32):
    """Rotate an integer right."""
    return (num >> shift) | (num << size - shift)



'''Cryptanalysis swiss army knife'''

__version__ = "0.6.0"

_author = 'Di Paola Martin'
_license = 'GNU GPLv3'
_url = ' - '

_license_disclaimer = r'''Copyright (C) {author} - {url}

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

try:
    from .conv import B, load_bytes
except ImportError:
    pass  # this happens when importing from setup.py



# https://packaging.python.org/en/latest/distributing.html
# https://github.com/pypa/sampleproject

from setuptools import setup, find_packages
from codecs import open
from os import path, system

import sys, re

here = path.abspath(path.dirname(__file__))

# load __version__, __doc__, _author, _license and _url
exec(open(path.join(here, 'cryptonita', '__init__.py')).read())

here = path.abspath(path.dirname(__file__))

try:
    system('''pandoc -f markdown-raw_html -o '%(dest_rst)s' '%(src_md)s' ''' % {
                'dest_rst': path.join(here, 'README.rst'),
                'src_md':   path.join(here, 'README.md'),
                })

    with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()

    # strip out any HTML comment|tag
    long_description = re.sub(r'<!--.*?-->', '', long_description,
                                                flags=re.DOTALL|re.MULTILINE)
    long_description = re.sub(r'<img.*?src=.*?>', '', long_description,
                                                flags=re.DOTALL|re.MULTILINE)

    with open(path.join(here, 'README.rst'), 'w', encoding='utf-8') as f:
        f.write(long_description)

except:
    print("Generation of the documentation failed. " + \
          "Do you have 'pandoc' installed?")

    long_description = __doc__

install_deps = [
        'base58',
    ]


extra_deps = {
        'full' : [
            'matplotlib',
            'numpy',
            'seaborn',
            'pycryptodome',
            'aspell-python-py3',    # apt-get install libaspell-dev
            'scipy',
            'langdetect',
            'gmpy2',       # apt-get install libgmp-dev libmpc-dev libmpfr-dev
            'z3-solver',
            ]
        }

setup(
    name='cryptonita',
    version=__version__,

    description=__doc__,
    long_description=long_description,

    #url=_url,

    # Author details
    author=_author,
    author_email='use-github-issues@example.com',

    license=_license,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',

        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        'Programming Language :: Python :: 3',
    ],

    # Let setuptools to find all the packages (aka modules and submodules)
    # automatically for us
    packages=find_packages(),
    python_requires='>=3.3',
    install_requires=install_deps,
    extras_require=extra_deps,

    keywords='crypto cryptography crypto-analysis',
)




from cryptonita.deps import importdep

np = importdep('numpy')
gp = importdep('gmpy2')
'''
>>> import numpy as np
>>> import gmpy2 as gp

>>> from cryptonita.mod import inv_matrix        # byexample: +timeout=10
'''


def inv_matrix(A, m):
    ''' Find the inverse of the matrix A module m if such exists.

        >>> A = [[2, 4,  5],
        ...      [9, 2,  1],
        ...      [3, 17, 7]]

        >>> inv_A = inv_matrix(A, m=26)

        >>> np.dot(A, inv_A) % 26
        array([[1, 0, 0],
               [0, 1, 0],
               [0, 0, 1]])
        '''

    det_A = np.linalg.det(A)
    inv_A = np.linalg.inv(A)

    # Adjugate matrix of A
    # https://en.wikipedia.org/wiki/Adjugate_matrix
    adj_A = det_A * inv_A

    tmp = np.round(det_A)
    #np.isclose(det_A, tmp)
    det_A = int(tmp)

    inv_det_A = int(gp.invert(det_A, m))

    inv_A = adj_A * inv_det_A
    inv_A = np.round(inv_A)
    return inv_A.astype(int) % m



'''
>>> from cryptonita.toys.hashes.md4 import md4
'''

# Pure python implementation of MD4, modified to be useful
# for cryptonita
#
# Src: https://gist.githubusercontent.com/kangtastic/c3349fc4f9d659ee362b12d7d8c639b6/raw/3b3317e2987c68f21fb2393547b2b0593a265f14/md4.py
# License: WTFPL, version 2 (wtfpl.net).
# Copyright © 2019 James Seo <james@equiv.tech> (github.com/kangtastic).
import struct


def md4(
    msg,
    h0=0x67452301,
    h1=0xEFCDAB89,
    h2=0x98BADCFE,
    h3=0x10325476,
    forged_message_len=None,
):
    '''
    >>> md4(b'')
    '31d6cfe0d16ae931b73c59d7e0c089c0'

    >>> md4(b'The quick brown fox jumps over the lazy dog')
    '1bee69a46ba811185c194762abaeae90'

    >>> md4(b'BEES')
    '501af1ef4b68495b5b7e37b15b4cda68'
    '''

    msg = bytes(msg)

    mask = 0xFFFFFFFF

    # Pre-processing: Total length is a multiple of 512 bits.
    ml = len(msg) * 8
    msg += b"\x80"
    msg += b"\x00" * (-(len(msg) + 8) % 64)
    if forged_message_len is None:
        msg += struct.pack("<Q", ml)
    else:
        msg += struct.pack("<Q", forged_message_len * 8)

    # Process the message in successive 512-bit chunks.
    chunks = [msg[i:i + 64] for i in range(0, len(msg), 64)]

    for chunk in chunks:
        X, h = list(struct.unpack("<16I", chunk)), [h0, h1, h2, h3]

        # Round 1.
        Xi = [3, 7, 11, 19]
        for n in range(16):
            i, j, k, l = map(lambda x: x % 4, range(-n, -n + 4))
            K, S = n, Xi[n % 4]
            hn = h[i] + F(h[j], h[k], h[l]) + X[K]
            h[i] = lrot32(hn & mask, S)

        # Round 2.
        Xi = [3, 5, 9, 13]
        for n in range(16):
            i, j, k, l = map(lambda x: x % 4, range(-n, -n + 4))
            K, S = n % 4 * 4 + n // 4, Xi[n % 4]
            hn = h[i] + G(h[j], h[k], h[l]) + X[K] + 0x5A827999
            h[i] = lrot32(hn & mask, S)

        # Round 3.
        Xi = [3, 9, 11, 15]
        Ki = [0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15]
        for n in range(16):
            i, j, k, l = map(lambda x: x % 4, range(-n, -n + 4))
            K, S = Ki[n], Xi[n % 4]
            hn = h[i] + H(h[j], h[k], h[l]) + X[K] + 0x6ED9EBA1
            h[i] = lrot32(hn & mask, S)

        h0, h1, h2, h3 = [
            ((v + n) & mask) for v, n in zip([h0, h1, h2, h3], h)
        ]

    return "".join(
        f"{value:02x}" for value in struct.pack("<4L", *[h0, h1, h2, h3])
    )


def F(x, y, z):
    return (x & y) | (~x & z)


def G(x, y, z):
    return (x & y) | (x & z) | (y & z)


def H(x, y, z):
    return x ^ y ^ z


def lrot32(value, n):
    lbits, rbits = (value << n) & 0xFFFFFFFF, value >> (32 - n)
    return lbits | rbits


def main():
    # Import is intentionally delayed.
    import sys

    if len(sys.argv) > 1:
        messages = [msg.encode() for msg in sys.argv[1:]]
        for message in messages:
            print(MD4(message).hexdigest())
    else:
        messages = [
            b"", b"The quick brown fox jumps over the lazy dog", b"BEES"
        ]
        known_hashes = [
            "31d6cfe0d16ae931b73c59d7e0c089c0",
            "1bee69a46ba811185c194762abaeae90",
            "501af1ef4b68495b5b7e37b15b4cda68",
        ]

        print("Testing the MD4 class.")
        print()

        for message, expected in zip(messages, known_hashes):
            print("Message: ", message)
            print("Expected:", expected)
            print("Actual:  ", MD4(message).hexdigest())
            print()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass
